# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Simon Heybrock
from typing import Dict, List, Optional, Union

import numpy as np
import scipp as sc

from .typing import ScippIndex


def convert_time_to_datetime64(
    raw_times: sc.Variable,
    start: str = None,
    scaling_factor: Union[float, np.float_] = None,
) -> sc.Variable:
    """
    The nexus standard allows an arbitrary scaling factor to be inserted
    between the numbers in the `time` series and the unit of time reported
    in the nexus attribute.

    The times are also relative to a given log start time, which might be
    different for each log. If this log start time is not available, the start of the
    unix epoch (1970-01-01T00:00:00) is used instead.

    See https://manual.nexusformat.org/classes/base_classes/NXlog.html

    Args:
        raw_times: The raw time data from a nexus file.
        start: Optional, the start time of the log in an ISO8601
            string. If not provided, defaults to the beginning of the
            unix epoch (1970-01-01T00:00:00).
        scaling_factor: Optional, the scaling factor between the provided
            time series data and the unit of the raw_times Variable. If
            not provided, defaults to 1 (a no-op scaling factor).
    """
    if (
        raw_times.dtype in (sc.DType.float64, sc.DType.float32)
    ) or scaling_factor is not None:
        unit = sc.units.ns
    else:
        # determine more precise unit
        ratio = sc.scalar(1.0, unit=start.unit) / sc.scalar(
            1.0, unit=raw_times.unit
        ).to(unit=start.unit)
        unit = start.unit if ratio.value < 1.0 else raw_times.unit

    if scaling_factor is None:
        times = raw_times
    else:
        times = raw_times * sc.scalar(value=scaling_factor)
    return start.to(unit=unit, copy=False) + times.to(
        dtype=sc.DType.int64, unit=unit, copy=False
    )


def _to_canonical_select(
    dims: List[str], select: ScippIndex
) -> Dict[str, Union[int, slice]]:
    """Return selection as dict with explicit dim labels"""

    def check_1d():
        if len(dims) != 1:
            raise sc.DimensionError(
                f"Dataset has multiple dimensions {dims}, "
                "specify the dimension to index."
            )

    if select is Ellipsis:
        return {}
    if isinstance(select, tuple) and len(select) == 0:
        return {}
    if isinstance(select, tuple) and isinstance(select[0], str):
        key, sel = select
        return {key: sel}
    if isinstance(select, tuple):
        check_1d()
        if len(select) != 1:
            raise sc.DimensionError(
                f"Dataset has single dimension {dims}, "
                "but multiple indices {select} were specified."
            )
        return {dims[0]: select[0]}
    elif isinstance(select, int) or isinstance(select, slice):
        check_1d()
        return {dims[0]: select}
    if not isinstance(select, dict):
        raise IndexError(f"Cannot process index {select}.")
    return select.copy()


def to_plain_index(dims: List[str], select: ScippIndex) -> Union[int, slice, tuple]:
    """
    Given a valid "scipp" index 'select', return an equivalent plain numpy-style index.
    """
    select = _to_canonical_select(dims, select)
    index = [slice(None)] * len(dims)
    for key, sel in select.items():
        if key not in dims:
            raise sc.DimensionError(
                f"'{key}' used for indexing not found in dataset dims {dims}."
            )
        index[dims.index(key)] = sel
    if len(index) == 1:
        return index[0]
    return tuple(index)


def to_child_select(
    dims: List[str],
    child_dims: List[str],
    select: ScippIndex,
    bin_edge_dim: Optional[str] = None,
) -> ScippIndex:
    """
    Given a valid "scipp" index 'select' for a Nexus class, return a selection for a
    child field of the class, which may have fewer dimensions.

    This removes any selections that apply to the parent but not the child.
    """
    select = _to_canonical_select(dims, select)
    for d in dims:
        if d not in child_dims and d in select:
            del select[d]
    for dim in select:
        if dim == bin_edge_dim:
            index = select[dim]
            if isinstance(index, int):
                select[dim] = slice(index, index + 2)
            elif index.stop > index.start:
                select[dim] = slice(index.start, index.stop + 1)
    return select



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Simon Heybrock
from typing import Any, Callable, Dict, Literal, Optional, Union

import scipp as sc

from ...base import Group, NXobject, base_definitions, create_field
from ...field import Field
from ...nxdata import NXdata
from ...typing import H5Group


class SASentry:
    nx_class = 'NXentry'

    def __init__(self, *, title: str, run: Union[str, int]):
        self.title = title
        self.run = run

    def __write_to_nexus_group__(self, group: H5Group):
        group.attrs['canSAS_class'] = 'SASentry'
        group.attrs['version'] = '1.0'
        group.attrs['definition'] = 'NXcanSAS'
        create_field(group, 'title', self.title)
        create_field(group, 'run', self.run)


class SASdata:
    nx_class = 'NXdata'

    def __init__(
        self,
        data: sc.DataArray,
        Q_variances: Optional[Literal['uncertainties', 'resolutions']] = None,
    ):
        self.data = data
        valid = ('uncertainties', 'resolutions')
        if Q_variances not in (None,) + valid:
            raise ValueError(f"Q_variances must be in {valid}")
        self._variances = Q_variances

    def __write_to_nexus_group__(self, group: H5Group):
        da = self.data
        group.attrs['canSAS_class'] = 'SASdata'
        group.attrs['signal'] = 'I'
        group.attrs['axes'] = da.dims  # for NeXus compliance, same as I_axes
        group.attrs['I_axes'] = da.dims
        group.attrs['Q_indices'] = tuple(da.dims.index(d) for d in da.coords['Q'].dims)
        # TODO writing Field should deal with variances
        signal = create_field(group, 'I', sc.values(da.data))
        # We use the _errors suffix for NeXus compliance, unlike the examples given in
        # NXcanSAS.
        if da.variances is not None:
            signal.attrs['uncertainties'] = 'I_errors'
            create_field(group, 'I_errors', sc.stddevs(da.data))
        if da.coords.is_edges('Q'):
            raise ValueError(
                "Q is given as bin-edges, but NXcanSAS requires Q points (such as "
                "bin centers)."
            )
        coord = create_field(group, 'Q', da.coords['Q'])
        if da.coords['Q'].variances is not None:
            if self._variances is None:
                raise ValueError(
                    "Q has variances, must specify whether these represent "
                    "'uncertainties' or 'resolutions' using the 'Q_variances' option'"
                )

            coord.attrs[self._variances] = 'Q_errors'
            # The errors are written automatically by create_field.


class _SASdata(NXdata):
    def __init__(self, attrs: Dict[str, Any], children: Dict[str, Union[Field, Group]]):
        fallback_dims = attrs.get('I_axes')
        if fallback_dims is not None:
            fallback_dims = (fallback_dims,)
        super().__init__(
            attrs=attrs,
            children=children,
            fallback_dims=fallback_dims,
            fallback_signal_name='I',
        )

    # TODO Mechanism for custom error names
    @staticmethod
    def signal_errors(group: NXobject) -> Optional[str]:
        signal_name = group.attrs.get('signal', 'I')
        signal = group._group[signal_name]
        return signal.attrs.get('uncertainties')

    def coord_errors(group: NXobject, name: str) -> Optional[str]:
        if name != 'Q':
            return None
        # TODO This naively stores this as Scipp errors, which are just Gaussian.
        # This is probably not correct in all cases.
        uncertainties = group[name].attrs.get('uncertainties')
        resolutions = group[name].attrs.get('resolutions')
        if uncertainties is None:
            return resolutions
        elif resolutions is None:
            return uncertainties
        raise RuntimeError("Cannot handle both uncertainties and resolutions for Q")


class _SAStransmission_spectrum(NXdata):
    def __init__(self, attrs: Dict[str, Any], children: Dict[str, Union[Field, Group]]):
        # TODO A valid file should have T_axes, do we need to fallback?
        super().__init__(
            attrs=attrs,
            children=children,
            fallback_dims=(attrs.get('T_axes', 'lambda'),),
            fallback_signal_name='T',
        )


class NXcanSAS:
    def get(self, key: type, default: Callable) -> Callable:
        def _definition_factory(
            attrs: Dict[str, Any], children: Dict[str, Union[Field, Group]]
        ) -> NXobject:
            if (cls := attrs.get('canSAS_class')) is not None:
                if cls == 'SASdata':
                    return _SASdata(attrs, children)
                if cls == 'SAStransmission_spectrum':
                    return _SAStransmission_spectrum(attrs, children)
            return base_definitions().get(key, default)(attrs, children)

        return _definition_factory


definitions = NXcanSAS()



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

# flake8: noqa
import importlib.metadata

try:
    __version__ = importlib.metadata.version(__package__ or __name__)
except importlib.metadata.PackageNotFoundError:
    __version__ = "0.0.0"

from . import typing
from .base import (
    Group,
    NexusStructureError,
    NXobject,
    base_definitions,
    create_class,
    create_field,
)
from .field import Attrs, Field
from .file import File
from .nexus_classes import *
from .nxtransformations import compute_positions, zip_pixel_offsets



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Simon Heybrock

# flake8: noqa
import importlib.metadata

try:
    __version__ = importlib.metadata.version(__package__ or __name__)
except importlib.metadata.PackageNotFoundError:
    __version__ = "0.0.0"

from .. import typing
from .definition import ApplicationDefinition, make_definition
from .file import File
from .nexus_classes import *
from .nxdata import NXdataStrategy
from .nxdetector import NXdetectorStrategy
from .nxlog import NXlogStrategy
from .nxobject import Attrs, Field, NexusStructureError, NXobject, NXobjectStrategy



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Simon Heybrock

__all__ = ['get_path']


def _make_pooch():
    import pooch

    return pooch.create(
        path=pooch.os_cache('scippnexus-externalfile'),
        env='SCIPPNEXUS_DATA_DIR',
        retry_if_failed=3,
        base_url='login.esss.dk:/mnt/groupdata/scipp/testdata/scippnexus/',
        registry={
            '2023/DREAM_baseline_all_dets.nxs': 'md5:1e1a3141c6785d25777b456e2a653f42',
            '2023/BIFROST_873855_00000015.hdf': 'md5:eb180b09d265c308e81c4a4885662bbd',
            '2023/DREAM_mccode.h5': 'md5:ebe4be53f20e139e865bc20b264bdceb',
            '2023/LOKI_mcstas_nexus_geometry.nxs': 'md5:f431d9775a53caffeebe9b879189b17c',  # noqa: E501
            '2023/NMX_2e11-rechunk.h5': 'md5:1174c208614b2e7a5faddc284b41d2c9',
            '2023/YMIR_038243_00010244.hdf': 'md5:cefb04b6d4d36f16e7f329a6045ad129',
            '2023/amor2020n000346_tweaked.nxs': 'md5:4e07ccc87b5c6549e190bc372c298e83',
            '2023/LOKI_60322-2022-03-02_2205_fixed.nxs': 'md5:e174d87dff10cbdcc2a33d093ae1e8ce',  # noqa: E501
        },
    )


_pooch = _make_pooch()


def sshdownloader(url, output_file, pooch):
    from subprocess import call

    cmd = ['scp', f'{url}', f'{output_file}']
    call(cmd)


def get_path(name: str) -> str:
    """
    Get path of file "downloaded" via SSH from login.esss.dk.

    You must have setup SSH agent for passwordless login for this to work.
    """
    return _pooch.fetch(name, downloader=sshdownloader)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Simon Heybrock
from typing import Any, Dict, List, Tuple, Union

import numpy as np
import scipp as sc

from ._common import to_plain_index
from .base import (
    Group,
    NexusStructureError,
    NXobject,
    ScippIndex,
    base_definitions_dict,
)
from .field import Field

_event_dimension = "event"
_pulse_dimension = "event_time_zero"


def _check_for_missing_fields(fields):
    for field in NXevent_data.mandatory_fields:
        if field not in fields:
            raise NexusStructureError(
                f"Required field {field} not found in NXevent_data"
            )


class NXevent_data(NXobject):
    mandatory_fields = ("event_time_zero", "event_index", "event_time_offset")
    handled_fields = mandatory_fields + ("event_id",)

    def __init__(self, attrs: Dict[str, Any], children: Dict[str, Union[Field, Group]]):
        super().__init__(attrs=attrs, children=children)
        for name, field in children.items():
            if name in ['event_time_zero', 'event_index']:
                field.sizes = {_pulse_dimension: field.dataset.shape[0]}
            elif name in ['event_time_offset', 'event_id']:
                field.sizes = {_event_dimension: field.dataset.shape[0]}

    @property
    def shape(self) -> Tuple[int]:
        if (event_index := self._children.get('event_index')) is not None:
            return event_index.shape
        return ()

    @property
    def dims(self) -> List[str]:
        return (_pulse_dimension,)[: len(self.shape)]

    @property
    def sizes(self) -> Dict[str, int]:
        return dict(zip(self.dims, self.shape))

    def field_dims(self, name: str, field: Field) -> Tuple[str, ...]:
        if name in ['event_time_zero', 'event_index']:
            return (_pulse_dimension,)
        if name in ['event_time_offset', 'event_id']:
            return (_event_dimension,)
        return None

    def read_children(self, select: ScippIndex) -> sc.DataGroup:
        children = self._children
        if not children:  # TODO Check that select is trivial?
            return sc.DataGroup()

        index = to_plain_index([_pulse_dimension], select)
        event_time_zero = children['event_time_zero'][index]
        last_loaded, event_index = self._get_event_index(children, index)

        num_event = children["event_time_offset"].shape[0]
        # Some files contain uint64 "max" indices, which turn into negatives during
        # conversion to int64. This is a hack to get around this.
        event_index[event_index < 0] = num_event

        if len(event_index) > 0:
            event_select = slice(
                event_index[0], event_index[-1] if last_loaded else num_event
            )
        else:
            event_select = slice(0, 0)

        event_time_offset = children['event_time_offset'][event_select]

        event_index = sc.array(
            dims=[_pulse_dimension],
            values=event_index[:-1] if last_loaded else event_index,
            dtype=sc.DType.int64,
            unit=None,
        )

        event_index -= event_index.min()

        dg = sc.DataGroup(
            event_time_zero=event_time_zero,
            event_index=event_index,
            event_time_offset=event_time_offset,
        )
        if (event_id := children.get('event_id')) is not None:
            dg['event_id'] = event_id[event_select]
        return dg

    def _get_event_index(self, children: sc.DataGroup, index):
        max_index = self.shape[0]
        if index is Ellipsis or index == tuple():
            last_loaded = False
        else:
            if isinstance(index, int):
                start, stop, _ = slice(index, None).indices(max_index)
                if start == stop:
                    raise IndexError(f'Index {start} is out of range')
                index = slice(start, start + 1)
            start, stop, stride = index.indices(max_index)
            if stop + stride > max_index:
                last_loaded = False
            elif start == stop:
                last_loaded = True
            else:
                stop += stride
                last_loaded = True
            index = slice(start, stop, stride)

        event_index = children['event_index'][index].values
        return last_loaded, event_index

    def assemble(self, children: sc.DataGroup) -> sc.DataArray:
        _check_for_missing_fields(children)
        event_time_offset = children['event_time_offset']
        event_time_zero = children['event_time_zero']
        event_index = children['event_index']

        # Weights are not stored in NeXus, so use 1s
        weights = sc.ones(
            dims=[_event_dimension],
            shape=event_time_offset.shape,
            unit='counts',
            dtype=np.float32,
        )

        events = sc.DataArray(
            data=weights, coords={'event_time_offset': event_time_offset}
        )
        if (event_id := children.get('event_id')) is not None:
            events.coords['event_id'] = event_id

        # There is some variation in the last recorded event_index in files from
        # different institutions. We try to make sure here that it is what would be the
        # first index of the next pulse. In other words, ensure that event_index
        # includes the bin edge for the last pulse.
        if event_time_zero.ndim == 0:
            begins = event_index[_pulse_dimension, 0]
        else:
            begins = event_index

        try:
            binned = sc.bins(data=events, dim=_event_dimension, begin=begins)
        except IndexError as e:
            path = self._children['event_index'].name
            raise NexusStructureError(f"Invalid index in NXevent_data at {path}:\n{e}")

        return sc.DataArray(data=binned, coords={'event_time_zero': event_time_zero})


base_definitions_dict['NXevent_data'] = NXevent_data



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Simon Heybrock
from typing import Optional, Tuple, Union

import scipp as sc

from .nxobject import NexusStructureError, NXobject


def off_to_shape(
    *,
    vertices: sc.Variable,
    winding_order: sc.Variable,
    faces: sc.Variable,
    detector_faces: Optional[sc.Variable] = None,
    detector_number: Optional[sc.Variable] = None,
) -> sc.Variable:
    """
    Convert OFF shape description to simpler shape representation.
    """
    # Vertices in winding order. This duplicates vertices if they are part of more than
    # one faces.
    vw = vertices[winding_order.values]
    # Same as above, grouped by face.
    fvw = sc.bins(begin=faces, data=vw, dim=vw.dim)
    low = fvw.bins.size().min().value
    high = fvw.bins.size().max().value
    if low == high:
        # Vertices in winding order, groupbed by face. Unlike `fvw` above we now know
        # that each face has the same number of vertices, so we can fold instead of
        # using binned data.
        shapes = vw.fold(dim=vertices.dim, sizes={faces.dim: -1, vertices.dim: low})
    else:
        raise NotImplementedError(
            "Conversion from OFF to shape not implemented for "
            "inconsistent number of vertices in faces."
        )
    if detector_faces is None:  # if detector_number is not None, all have same shape
        return sc.bins(begin=sc.index(0), dim=faces.dim, data=shapes)
    if detector_number is None:
        raise NexusStructureError(
            "`detector_number` not given but NXoff_geometry "
            "contains `detector_faces`."
        )
    shape_index = detector_faces['column', 0].copy()
    detid = detector_faces['column', 1].copy()
    da = sc.DataArray(shape_index, coords={'detector_number': detid}).group(
        detector_number.flatten(to='detector_number')
    )
    comps = da.bins.constituents
    comps['data'] = shapes[faces.dim, comps['data'].values]
    return sc.bins(**comps).fold(dim='detector_number', sizes=detector_number.sizes)


class NXoff_geometry(NXobject):
    _dims = {
        'detector_faces': ('face', 'column'),
        'vertices': ('vertex',),
        'winding_order': ('winding_order',),
        'faces': ('face',),
    }

    def _get_field_dims(self, name: str) -> Union[None, Tuple[str]]:
        return self._dims.get(name)

    def _get_field_dtype(self, name: str) -> Union[None, sc.DType]:
        if name == 'vertices':
            return sc.DType.vector3
        return None

    def load_as_array(
        self, detector_number: Optional[sc.Variable] = None
    ) -> sc.Variable:
        return off_to_shape(**self[()], detector_number=detector_number)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Scipp contributors (https://github.com/scipp)

import argparse
import sys
from typing import List

import git
from packaging.version import InvalidVersion, Version, parse


def _get_releases() -> List[Version]:
    """Return reversed sorted list of release tag names."""
    tags = git.Repo(search_parent_directories=True).tags
    versions = []
    for t in tags:
        try:
            versions.append(parse(t.name))
        except InvalidVersion:
            pass
    return sorted(versions, reverse=True)


class VersionInfo:
    def __init__(self):
        self._releases = _get_releases()

    def _to_version(self, version) -> Version:
        if isinstance(version, str):
            try:
                return parse(version)
            except InvalidVersion:
                # When not building for a tagged release we may get, e.g., 'main'.
                # Pretend this means the current latest release.
                return self._releases[0]
        return version

    def minor_releases(self, first: str = '0.1') -> List[str]:
        """Return set of minor releases in the form '1.2'.

        `first` gives the first release to be included. By default '0.0' releases are
        not included.
        """
        first = parse(first)
        releases = [r for r in self._releases if r >= first]
        releases = sorted(set((r.major, r.minor) for r in releases), reverse=True)
        return [f'{major}.{minor}' for major, minor in releases]

    def is_latest(self, version: str) -> bool:
        """Return True if `version` has the same or larger major and minor as the
        latest release.
        """
        version = self._to_version(version)
        latest = self._releases[0]
        return (latest.major, latest.minor) <= (version.major, version.minor)

    def is_new(self, version: str) -> bool:
        """Return True if `version` is a new major or minor release."""
        version = self._to_version(version)
        releases = [r for r in self._releases if r != version]
        latest = releases[0]
        return (latest.major, latest.minor) < (version.major, version.minor)

    def target(self, version: str) -> str:
        version = self._to_version(version)
        if self.is_latest(version):
            return ''
        else:
            return f'release/{version.major}.{version.minor}'

    def replaced(self, version: str) -> str:
        version = self._to_version(version)
        # If we release 1.2 we want to find 1.1
        for release in self._releases:
            if (release.major, release.minor) < (version.major, version.minor):
                return release


def main(action: str, version: str) -> int:
    info = VersionInfo()
    if action == 'is-latest':
        print(info.is_latest(version))
    elif action == 'is-new':
        print(info.is_new(version))
    elif action == 'get-replaced':
        print(info.replaced(version))
    elif action == 'get-target':
        print(info.target(version))
    return 0


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Process some integers.')
    parser.add_argument(
        '--action',
        choices=['is-latest', 'is-new', 'get-replaced', 'get-target'],
        required=True,
        help='Action to perform: Check whether this major or minor '
        'release exists or is new (is-latest), check whether this is a '
        'new major or minor release (is-new), get the version this is '
        'replacing (get-replaced), get the target folder for '
        'publishing the docs (get-target). In all cases the '
        'patch/micro version is ignored.',
    )
    parser.add_argument(
        '--version', dest='version', required=True, help='Version the action refers to'
    )

    args = parser.parse_args()
    sys.exit(main(**vars(args)))



import h5py
import numpy as np
import pytest
import scipp as sc
from scipp import spatial
from scipp.testing import assert_identical

import scippnexus as snx


@pytest.fixture()
def nxroot():
    """Yield NXroot containing a single NXentry named 'entry'"""
    with h5py.File('dummy.nxs', mode='w', driver="core", backing_store=False) as f:
        root = snx.Group(f, definitions=snx.base_definitions())
        root.create_class('entry', snx.NXentry)
        yield root


def test_ub_matrix_loaded_as_linear_transform_with_inverse_angstrom_unit(nxroot):
    sample = nxroot.create_class('data1', snx.NXsample)
    matrix = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    sample['ub_matrix'] = matrix
    loaded = sample[()]
    assert_identical(
        loaded,
        sc.DataGroup(
            ub_matrix=spatial.linear_transform(value=matrix, unit='1/angstrom')
        ),
    )


def test_ub_matrix_array_can_be_loaded(nxroot):
    sample = nxroot.create_class('data1', snx.NXsample)
    matrices = np.array(
        [[[1, 2, 3], [4, 5, 6], [7, 8, 9]], [[2, 3, 4], [5, 6, 7], [8, 9, 10]]]
    )
    sample['ub_matrix'] = matrices
    loaded = sample[()]
    assert_identical(
        loaded,
        sc.DataGroup(
            ub_matrix=spatial.linear_transforms(
                dims=('dim_0',), values=matrices, unit='1/angstrom'
            )
        ),
    )


def test_orientation_matrix_loaded_as_linear_transform_with_dimensionless_unit(nxroot):
    sample = nxroot.create_class('data1', snx.NXsample)
    matrix = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    sample['orientation_matrix'] = matrix
    loaded = sample[()]
    assert_identical(
        loaded,
        sc.DataGroup(
            orientation_matrix=spatial.linear_transform(value=matrix, unit='')
        ),
    )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Simon Heybrock
from typing import Dict, Union

from .nxdetector import NXdetector
from .nxobject import Field


class NXmonitor(NXdetector):
    @property
    def _event_grouping(self) -> Dict[str, Union[str, Field]]:
        # Unlike NXdetector, NXmonitor does not group by 'detector_number'. We pass
        # grouping information that matches the underlying binning of NXevent_data
        # such that no addition binning will need to be performed. That is, the by-pulse
        # binning present in the file (in NXevent_data) is preserved.
        return {
            'grouping_key': 'event_time_zero',
            'grouping': self.events.get('event_time_zero'),
        }



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Simon Heybrock
import warnings
from contextlib import AbstractContextManager

import h5py
from scipp import VisibleDeprecationWarning

from .nexus_classes import NXroot


class File(AbstractContextManager, NXroot):
    def __init__(self, *args, definition=None, **kwargs):
        warnings.warn(
            "The scippnexus.v1 API is deprecated and will be removed in release 24.01. "
            "Switch to 'import scippnexus as snx' to use API version 2.",
            VisibleDeprecationWarning,
            stacklevel=2,
        )
        self._file = h5py.File(*args, **kwargs)
        NXroot.__init__(self, self._file, definition=definition)

    def __enter__(self):
        self._file.__enter__()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self._file.close()

    def close(self):
        self._file.close()



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Simon Heybrock
from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Dict, List, Protocol, Tuple, Union


class H5Base(Protocol):
    @property
    def attrs(self) -> Dict:
        """Attributes of dataset or group"""

    @property
    def name(self) -> str:
        """Name of dataset or group"""

    @property
    def file(self) -> List[int]:
        """File of dataset or group"""

    @property
    def parent(self) -> H5Group:
        """Parent of dataset or group"""


# TODO Define more required methods
class H5Dataset(H5Base, Protocol):
    """h5py.Dataset-like"""

    @property
    def shape(self) -> List[int]:
        """Shape of a dataset"""

    @property
    def dtype(self) -> List[int]:
        """dtype of a dataset"""

    def read_direct(self, array) -> None:
        """Read dataset into given buffer"""


class H5Group(H5Base, Protocol):
    """h5py.Group-like"""

    def __getitem__(self, index: Union[str, Any]) -> Union[H5Dataset, H5Group]:
        """Keys in the group"""

    def keys(self) -> List[str]:
        """Keys in the group"""

    def create_dataset(self) -> H5Dataset:
        """Create a dataset"""

    def create_group(self) -> H5Group:
        """Create a group"""

    def visititems(self, func: Callable) -> None:
        """Apply callable to all items, recursively"""


if TYPE_CHECKING:
    from enum import Enum

    class ellipsis(Enum):
        Ellipsis = "..."

else:
    ellipsis = type(Ellipsis)

# Note that scipp does not support dicts yet, but this HDF5 code does, to
# allow for loading blocks of 2d (or higher) data efficiently.
ScippIndex = Union[
    ellipsis,
    int,
    tuple,
    slice,
    Tuple[str, Union[int, slice]],
    Dict[str, Union[int, slice]],
]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Simon Heybrock
from __future__ import annotations

from copy import copy
from typing import Any, Dict, List, Optional, Union

import scipp as sc

from .nxdata import NXdata, NXdataStrategy
from .nxevent_data import NXevent_data
from .nxobject import (
    Field,
    NexusStructureError,
    NXobject,
    ScippIndex,
    asarray,
    is_dataset,
)


class NXdetectorStrategy(NXdataStrategy):
    @staticmethod
    def signal(group):
        # NXdata uses the 'signal' attribute to define the field name of the signal.
        # NXdetector uses a "hard-coded" signal name 'data', without specifying the
        # attribute in the file, so we pass this explicitly to NXdata.
        # Note the special case of an NXevent_data subgroup named 'data', which we
        # avoid by checking if 'data' is a dataset.
        if (name := NXdataStrategy.signal(group)) is not None:
            return name
        return 'data' if 'data' in group and is_dataset(group._group['data']) else None


def group(da: sc.DataArray, groups: sc.Variable) -> sc.DataArray:
    if hasattr(da, 'group'):
        return da.group(groups)
    else:
        return sc.bin(da, groups=[groups])


class EventSelector:
    """A proxy object for creating an NXdetector based on a selection of events."""

    def __init__(self, detector):
        self._detector = detector

    def __getitem__(self, select: ScippIndex) -> NXdetector:
        """Return an NXdetector based on a selection (slice) of events."""
        det = copy(self._detector)
        det._event_select = select
        return det


class _EventField:
    """Field-like wrapper of NXevent_data binned into pixels.

    This has no equivalent in the NeXus format, but represents the conceptual
    event-data "signal" dataset of an NXdetector.
    """

    def __init__(
        self,
        nxevent_data: NXevent_data,
        event_select: ScippIndex,
        grouping_key: Optional[str] = 'detector_number',
        grouping: Optional[Field] = None,
    ):
        self._nxevent_data = nxevent_data
        self._event_select = event_select
        self._grouping_key = grouping_key
        self._grouping = grouping

    @property
    def name(self) -> str:
        return self._nxevent_data.name

    @property
    def attrs(self):
        return self._nxevent_data.attrs

    @property
    def dims(self):
        if self._grouping is None:
            return [self._grouping_key]
        return self._grouping.dims

    @property
    def shape(self):
        if self._grouping is None:
            raise NexusStructureError(
                "Cannot get shape of NXdetector since no 'detector_number' "
                "field found but detector contains event data."
            )
        return self._grouping.shape

    @property
    def unit(self) -> None:
        return self._nxevent_data.unit

    def __getitem__(self, select: ScippIndex) -> sc.DataArray:
        event_data = self._nxevent_data[self._event_select]
        if isinstance(event_data, sc.DataGroup):
            raise NexusStructureError("Invalid NXevent_data in NXdetector.")
        if self._grouping is None:
            if select not in (Ellipsis, tuple(), slice(None)):
                raise NexusStructureError(
                    "Cannot load slice of NXdetector since it contains event data "
                    "but no 'detector_number' field, i.e., the shape is unknown. "
                    "Use ellipsis or an empty tuple to load the full detector."
                )
            # Ideally we would prefer to use np.unique, but a quick experiment shows
            # that this can easily be 100x slower, so it is not an option. In
            # practice most files have contiguous event_id values within a bank
            # (NXevent_data).
            id_min = event_data.bins.coords['event_id'].min()
            id_max = event_data.bins.coords['event_id'].max()
            grouping = sc.arange(
                dim=self._grouping_key,
                unit=None,
                start=id_min.value,
                stop=id_max.value + 1,
                dtype=id_min.dtype,
            )
        else:
            grouping = asarray(self._grouping[select])
            if (self._grouping_key in event_data.coords) and sc.identical(
                grouping, event_data.coords[self._grouping_key]
            ):
                return event_data
        # copy since sc.bin cannot deal with a non-contiguous view
        event_id = grouping.flatten(to='event_id').copy()
        event_data.bins.coords['event_time_zero'] = sc.bins_like(
            event_data, fill_value=event_data.coords['event_time_zero']
        )
        # After loading raw NXevent_data it is guaranteed that the event table
        # is contiguous and that there is no masking. We can therefore use the
        # more efficient approach of binning from scratch instead of erasing the
        # 'pulse' binning defined by NXevent_data.
        event_data = group(event_data.bins.constituents['data'], groups=event_id)
        if self._grouping is None:
            event_data.coords[self._grouping_key] = event_data.coords.pop('event_id')
        else:
            del event_data.coords['event_id']
        return event_data.fold(dim='event_id', sizes=grouping.sizes)


class NXdetector(NXobject):
    """A detector or detector bank providing an array of values or events.

    If the detector stores event data then the 'detector_number' field (if present)
    is used to map events to detector pixels. Otherwise this returns event data in the
    same format as NXevent_data.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._event_select = tuple()
        self._nxevent_data_fields = [
            'event_time_zero',
            'event_index',
            'event_time_offset',
            'event_id',
            'cue_timestamp_zero',
            'cue_index',
            'pulse_height',
        ]
        self._detector_number_fields = ['detector_number', 'pixel_id', 'spectrum_index']

    @property
    def shape(self) -> List[int]:
        return self._signal.shape

    @property
    def dims(self) -> List[str]:
        return self._signal.dims

    @property
    def unit(self) -> Union[sc.Unit, None]:
        return self._signal.unit

    @property
    def detector_number(self) -> Optional[Field]:
        for key in self._detector_number_fields:
            if key in self:
                return key
        return None

    @property
    def _event_grouping(self) -> Union[None, Dict[str, Any]]:
        for key in self._detector_number_fields:
            if key in self:
                return {'grouping_key': key, 'grouping': self[key]}
        return {}

    @property
    def _signal(self) -> Union[Field, _EventField]:
        return self._nxdata()._signal

    def _nxdata(self, use_event_signal=True) -> NXdata:
        events = self.events
        if use_event_signal and events is not None:
            signal = _EventField(events, self._event_select, **self._event_grouping)
        else:
            signal = None
        skip = None
        if events is not None:
            if events.name == self.name:
                skip = self._nxevent_data_fields
            else:
                skip = [events.name.split('/')[-1]]  # name of the subgroup
        return NXdata(
            self._group, strategy=NXdetectorStrategy, signal_override=signal, skip=skip
        )

    @property
    def events(self) -> Union[None, NXevent_data]:
        """Return the underlying NXevent_data group, None if not event data."""
        # The standard is unclear on whether the 'data' field may be NXevent_data or
        # whether the fields of NXevent_data should be stored directly within this
        # NXdetector. Both cases are observed in the wild.
        event_entries = self[NXevent_data]
        if len(event_entries) > 1:
            raise NexusStructureError(
                "No unique NXevent_data entry in NXdetector. "
                f"Found {len(event_entries)}."
            )
        if len(event_entries) == 1:
            # If there is also a signal dataset (not events) it will be ignored
            # (except for possibly using it to deduce shape and dims).
            return next(iter(event_entries.values()))
        if 'event_time_offset' in self:
            return NXevent_data(self._group)
        return None

    @property
    def select_events(self) -> EventSelector:
        """
        Return a proxy object for selecting a slice of the underlying NXevent_data
        group, while keeping wrapping the NXdetector.
        """
        if self.events is None:
            raise NexusStructureError(
                "Cannot select events in NXdetector not containing NXevent_data."
            )
        return EventSelector(self)

    def _get_field_dims(self, name: str) -> Union[None, List[str]]:
        if self.events is not None:
            if name in self._nxevent_data_fields:
                # Event field is direct child of this class
                return self.events._get_field_dims(name)
            if name in self._detector_number_fields:
                # If there is a signal field in addition to the event data it can be
                # used to define dimension labels
                nxdata = self._nxdata(use_event_signal=False)
                if nxdata._signal_name is not None:
                    return nxdata._get_field_dims(name)
                # If grouping is 1-D then we use this name as the dim
                if self._get_child(name).ndim == 1:
                    return [name]
                return None
        return self._nxdata()._get_field_dims(name)

    def _getitem(self, select: ScippIndex) -> sc.DataArray:
        return self._nxdata()._getitem(select)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Simon Heybrock
from .leaf import Leaf


class NXdisk_chopper(Leaf):
    """Disk chopper information, can be read as a dict."""



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Simon Heybrock
from contextlib import AbstractContextManager
from typing import Mapping

import h5py

from .base import Group, base_definitions

_default_definitions = object()


class File(AbstractContextManager, Group):
    def __init__(self, *args, definitions: Mapping = _default_definitions, **kwargs):
        """Context manager for NeXus files, similar to h5py.File.

        Arguments other than documented are as in :py:class:`h5py.File`.

        Parameters
        ----------
        definitions:
            Mapping of NX_class names to application-specific definitions.
            The default is to use the base definitions as defined in the
            NeXus standard.
        """
        if definitions is _default_definitions:
            definitions = base_definitions()
        self._file = h5py.File(*args, **kwargs)
        super().__init__(self._file, definitions=definitions)

    def __enter__(self):
        self._file.__enter__()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self._file.close()

    def close(self):
        self._file.close()



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Simon Heybrock
import datetime
import posixpath
import re
import warnings
from dataclasses import dataclass
from types import MappingProxyType
from typing import TYPE_CHECKING, Any, Dict, Optional, Tuple, Union

import dateutil.parser
import numpy as np
import scipp as sc

from scippnexus._common import convert_time_to_datetime64, to_plain_index
from scippnexus._hdf5_nexus import _warn_latin1_decode
from scippnexus.typing import H5Dataset, ScippIndex

from ._cache import cached_property
from .attrs import Attrs

if TYPE_CHECKING:
    from .base import Group


def depends_on_to_relative_path(depends_on: str, parent_path: str) -> str:
    """Replace depends_on paths with relative paths.

    After loading we will generally not have the same root so absolute paths
    cannot be resolved after loading."""
    if depends_on.startswith('/'):
        return posixpath.relpath(depends_on, parent_path)
    return depends_on


def _is_time(obj):
    if (unit := obj.unit) is None:
        return False
    return unit.to_dict().get('powers') == {'s': 1}


def _as_datetime(obj: Any):
    if isinstance(obj, str):
        try:
            # NumPy and scipp cannot handle timezone information. We therefore apply it,
            # i.e., convert to UTC.
            # Would like to use dateutil directly, but with Python's datetime we do not
            # get nanosecond precision. Therefore we combine numpy and dateutil parsing.
            date_only = 'T' not in obj
            if date_only:
                return sc.datetime(obj)
            date, time = obj.split('T')
            time_and_timezone_offset = re.split(r'Z|\+|-', time)
            time = time_and_timezone_offset[0]
            if len(time_and_timezone_offset) == 1:
                # No timezone, parse directly (scipp based on numpy)
                return sc.datetime(obj)
            else:
                # There is timezone info. Parse with dateutil.
                dt = dateutil.parser.isoparse(obj)
                dt = dt.replace(microsecond=0)  # handled by numpy
                dt = dt.astimezone(datetime.timezone.utc)
                dt = dt.replace(tzinfo=None).isoformat()
                # We operate with string operations here and thus end up parsing date
                # and time twice. The reason is that the timezone-offset arithmetic
                # cannot be done, e.g., in nanoseconds without causing rounding errors.
                if '.' in time:
                    dt += f".{time.split('.')[1]}"
                return sc.datetime(dt)
        except ValueError:
            pass
    return None


@dataclass
class Field:
    """NeXus field.
    In HDF5 fields are represented as dataset.
    """

    dataset: H5Dataset
    parent: 'Group'
    sizes: Optional[Dict[str, int]] = None
    dtype: Optional[sc.DType] = None
    errors: Optional[H5Dataset] = None

    @cached_property
    def attrs(self) -> Dict[str, Any]:
        """The attributes of the dataset.
        Cannot be used for writing attributes, since they are cached for performance."""
        return MappingProxyType(Attrs(self.dataset.attrs))

    @property
    def dims(self) -> Tuple[str, ...]:
        return tuple(self.sizes.keys())

    @property
    def shape(self) -> Tuple[int, ...]:
        return tuple(self.sizes.values())

    @cached_property
    def file(self) -> 'Group':
        return self.parent.file

    def _load_variances(self, var, index):
        stddevs = sc.empty(
            dims=var.dims, shape=var.shape, dtype=var.dtype, unit=var.unit
        )
        try:
            self.errors.read_direct(stddevs.values, source_sel=index)
        except TypeError:
            stddevs.values = self.errors[index].squeeze()
        # According to the standard, errors must have the same shape as the data.
        # This is not the case in all files we observed, is there any harm in
        # attempting a broadcast?
        var.variances = np.broadcast_to(
            sc.pow(stddevs, sc.scalar(2)).values, shape=var.shape
        )

    def __getitem__(self, select: ScippIndex) -> Union[Any, sc.Variable]:
        """
        Load the field as a :py:class:`scipp.Variable` or Python object.

        If the shape is empty and no unit is given this returns a Python object, such
        as a string or integer. Otherwise a :py:class:`scipp.Variable` is returned.

        Parameters
        ----------
        select:
            Scipp-style index: Load the specified slice of the current group.

        Returns
        -------
        :
            Loaded data.
        """
        from .nxtransformations import maybe_transformation

        index = to_plain_index(self.dims, select)
        if isinstance(index, (int, slice)):
            index = (index,)

        base_dims = self.dims
        base_shape = self.shape
        dims = []
        shape = []
        for i, ind in enumerate(index):
            if not isinstance(ind, int):
                dims.append(base_dims[i])
                shape.append(len(range(*ind.indices(base_shape[i]))))

        variable = sc.empty(
            dims=dims,
            shape=shape,
            dtype=self.dtype,
            unit=self.unit,
            with_variances=self.errors is not None,
        )

        # If the variable is empty, return early
        if np.prod(shape) == 0:
            variable = self._maybe_datetime(variable)
            return maybe_transformation(self, value=variable, sel=select)

        if self.dtype == sc.DType.string:
            try:
                strings = self.dataset.asstr()[index]
            except UnicodeDecodeError as e:
                strings = self.dataset.asstr(encoding='latin-1')[index]
                _warn_latin1_decode(self.dataset, strings, str(e))
            variable.values = np.asarray(strings).flatten()
            if self.dataset.name.endswith('depends_on') and variable.ndim == 0:
                variable.value = depends_on_to_relative_path(
                    variable.value, self.dataset.parent.name
                )
        elif variable.values.flags["C_CONTIGUOUS"]:
            # On versions of h5py prior to 3.2, a TypeError occurs in some cases
            # where h5py cannot broadcast data with e.g. shape (20, 1) to a buffer
            # of shape (20,). Note that broadcasting (1, 20) -> (20,) does work
            # (see https://github.com/h5py/h5py/pull/1796).
            # Therefore, we manually squeeze here.
            # A pin of h5py<3.2 is currently required by Mantid and hence scippneutron
            # (see https://github.com/h5py/h5py/issues/1880#issuecomment-823223154)
            # hence this workaround. Once we can use a more recent h5py with Mantid,
            # this try/except can be removed.
            try:
                self.dataset.read_direct(variable.values, source_sel=index)
            except TypeError:
                variable.values = self.dataset[index].squeeze()
            if self.errors is not None:
                self._load_variances(variable, index)
        else:
            variable.values = self.dataset[index]
        if variable.ndim == 0 and variable.unit is None and variable.fields is None:
            # Work around scipp/scipp#2815, and avoid returning NumPy bool
            if isinstance(variable.values, np.ndarray) and variable.dtype != 'bool':
                return variable.values[()]
            else:
                return variable.value
        variable = self._maybe_datetime(variable)
        return maybe_transformation(self, value=variable, sel=select)

    def _maybe_datetime(self, variable: sc.Variable) -> sc.Variable:
        if _is_time(variable):
            starts = []
            for name in self.attrs:
                if (dt := _as_datetime(self.attrs[name])) is not None:
                    starts.append(dt)
            if len(starts) == 1:
                variable = convert_time_to_datetime64(
                    variable,
                    start=starts[0],
                    scaling_factor=self.attrs.get('scaling_factor'),
                )

        return variable

    def __repr__(self) -> str:
        return f'<Nexus field "{self.dataset.name}">'

    @property
    def name(self) -> str:
        return self.dataset.name

    @property
    def ndim(self) -> int:
        """Total number of dimensions in the dataset.
        See the shape property for potential differences to the value returned by the
        underlying h5py.Dataset.ndim.
        """
        return len(self.shape)

    @cached_property
    def unit(self) -> Union[sc.Unit, None]:
        if (unit := self.attrs.get('units')) is not None:
            try:
                return sc.Unit(unit)
            except sc.UnitError:
                warnings.warn(
                    f"Unrecognized unit '{unit}' for value dataset "
                    f"in '{self.name}'; setting unit as 'dimensionless'",
                    stacklevel=2,
                )
                return sc.units.one
        return None



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Simon Heybrock
from __future__ import annotations

import warnings
from dataclasses import dataclass
from typing import Dict, List, Optional, Tuple, Union

import numpy as np
import scipp as sc
from scipp.scipy import interpolate

from .base import Group, NexusStructureError, NXobject, ScippIndex
from .field import Field, depends_on_to_relative_path


class TransformationError(NexusStructureError):
    pass


class NXtransformations(NXobject):
    """Group of transformations."""


class Transformation:
    def __init__(self, obj: Union[Field, Group]):  # could be an NXlog
        self._obj = obj

    @property
    def sizes(self) -> dict:
        return self._obj.sizes

    @property
    def dims(self) -> Tuple[str, ...]:
        return self._obj.dims

    @property
    def shape(self) -> Tuple[int, ...]:
        return self._obj.shape

    @property
    def attrs(self):
        return self._obj.attrs

    @property
    def name(self):
        return self._obj.name

    @property
    def offset(self):
        if (offset := self.attrs.get('offset')) is None:
            return None
        if (offset_units := self.attrs.get('offset_units')) is None:
            raise TransformationError(
                f"Found {offset=} but no corresponding 'offset_units' "
                f"attribute at {self.name}"
            )
        return sc.spatial.translation(value=offset, unit=offset_units)

    @property
    def vector(self) -> sc.Variable:
        return sc.vector(value=self.attrs.get('vector'))

    def __getitem__(self, select: ScippIndex):
        transformation_type = self.attrs.get('transformation_type')
        # According to private communication with Tobias Richter, NeXus allows 0-D or
        # shape=[1] for single values. It is unclear how and if this could be
        # distinguished from a scan of length 1.
        value = self._obj[select]
        return self.make_transformation(value, transformation_type, select)

    def make_transformation(
        self,
        value: Union[sc.Variable, sc.DataArray],
        transformation_type: str,
        select: ScippIndex,
    ):
        try:
            if isinstance(value, sc.DataGroup):
                return value
            t = value * self.vector
            v = t if isinstance(t, sc.Variable) else t.data
            if transformation_type == 'translation':
                v = sc.spatial.translations(dims=v.dims, values=v.values, unit=v.unit)
            elif transformation_type == 'rotation':
                v = sc.spatial.rotations_from_rotvecs(v)
            else:
                raise TransformationError(
                    f"{transformation_type=} attribute at {self.name},"
                    " expected 'translation' or 'rotation'."
                )
            if isinstance(t, sc.Variable):
                t = v
            else:
                t.data = v
            if (offset := self.offset) is None:
                transform = t
            else:
                offset = sc.vector(value=offset.values, unit=offset.unit)
                offset = sc.spatial.translation(value=offset.value, unit=offset.unit)
                if transformation_type == 'translation':
                    offset = offset.to(unit=t.unit, copy=False)
                transform = t * offset
            if (depends_on := self.attrs.get('depends_on')) is not None:
                if not isinstance(transform, sc.DataArray):
                    transform = sc.DataArray(transform)
                transform.coords['depends_on'] = sc.scalar(
                    depends_on_to_relative_path(depends_on, self._obj.parent.name)
                )
            return transform
        except (sc.DimensionError, sc.UnitError, TransformationError):
            # TODO We should probably try to return some other data structure and
            # also insert offset and other attributes.
            return value


def _interpolate_transform(transform, xnew):
    # scipy can't interpolate with a single value
    if transform.sizes["time"] == 1:
        transform = sc.concat([transform, transform], dim="time")
    return interpolate.interp1d(
        transform, "time", kind="previous", fill_value="extrapolate"
    )(xnew=xnew)


def _smaller_unit(a, b):
    if a.unit == b.unit:
        return a.unit
    ratio = sc.scalar(1.0, unit=a.unit).to(unit=b.unit)
    if ratio.value < 1.0:
        return a.unit
    else:
        return b.unit


def combine_transformations(
    chain: List[Union[sc.DataArray, sc.Variable]]
) -> Union[sc.DataArray, sc.Variable]:
    """
    Take the product of a chain of transformations, handling potentially mismatching
    time-dependence.

    Time-dependent transformations are interpolated to a common time-coordinate.
    """
    total_transform = None
    for transform in chain:
        if total_transform is None:
            total_transform = transform
        elif isinstance(total_transform, sc.DataArray) and isinstance(
            transform, sc.DataArray
        ):
            unit = _smaller_unit(
                transform.coords['time'], total_transform.coords['time']
            )
            total_transform.coords['time'] = total_transform.coords['time'].to(
                unit=unit, copy=False
            )
            transform.coords['time'] = transform.coords['time'].to(
                unit=unit, copy=False
            )
            time = sc.concat(
                [total_transform.coords["time"], transform.coords["time"]], dim="time"
            )
            time = sc.datetimes(values=np.unique(time.values), dims=["time"], unit=unit)
            total_transform = _interpolate_transform(
                transform, time
            ) * _interpolate_transform(total_transform, time)
        else:
            total_transform = transform * total_transform
    if isinstance(total_transform, sc.DataArray):
        time_dependent = [t for t in chain if isinstance(t, sc.DataArray)]
        times = [da.coords['time'][0] for da in time_dependent]
        latest_log_start = sc.reduce(times).max()
        return total_transform['time', latest_log_start:].copy()
    return sc.scalar(1) if total_transform is None else total_transform


def maybe_transformation(
    obj: Union[Field, Group],
    value: Union[sc.Variable, sc.DataArray, sc.DataGroup],
    sel: ScippIndex,
) -> Union[sc.Variable, sc.DataArray, sc.DataGroup]:
    """
    Return a loaded field, possibly modified if it is a transformation.

    Transformations are usually stored in NXtransformations groups. However, identifying
    transformation fields in this way requires inspecting the parent group, which
    is cumbersome to implement. Furthermore, according to the NXdetector documentation
    transformations are not necessarily placed inside NXtransformations.
    Instead we use the presence of the attribute 'transformation_type' to identify
    transformation fields.
    """
    if (transformation_type := obj.attrs.get('transformation_type')) is not None:
        return Transformation(obj).make_transformation(value, transformation_type, sel)
    return value


class TransformationChainResolver:
    """
    Resolve a chain of transformations, given depends_on attributes with absolute or
    relative paths.

    A `depends_on` field serves as an entry point into a chain of transformations.
    It points to another entry, based on an absolute or relative path. The target
    entry may have a `depends_on` attribute pointing to the next transform. This
    class follows the paths and resolves the chain of transformations.
    """

    class ChainError(KeyError):
        """Raised when a transformation chain cannot be resolved."""

        pass

    @dataclass
    class Entry:
        name: str
        value: sc.DataGroup

    def __init__(self, stack: List[TransformationChainResolver.Entry]):
        self._stack = stack

    @staticmethod
    def from_root(dg: sc.DataGroup) -> TransformationChainResolver:
        return TransformationChainResolver(
            [TransformationChainResolver.Entry(name='', value=dg)]
        )

    @property
    def name(self) -> str:
        return '/'.join([e.name for e in self._stack])

    @property
    def root(self) -> TransformationChainResolver:
        return TransformationChainResolver(self._stack[0:1])

    @property
    def parent(self) -> TransformationChainResolver:
        if len(self._stack) == 1:
            raise TransformationChainResolver.ChainError(
                "Transformation depends on node beyond root"
            )
        return TransformationChainResolver(self._stack[:-1])

    @property
    def value(self) -> sc.DataGroup:
        return self._stack[-1].value

    def __getitem__(self, path: str) -> TransformationChainResolver:
        base, *remainder = path.split('/', maxsplit=1)
        if base == '':
            node = self.root
        elif base == '.':
            node = self
        elif base == '..':
            node = self.parent
        else:
            try:
                child = self._stack[-1].value[base]
            except KeyError:
                raise TransformationChainResolver.ChainError(
                    f"{base} not found in {self.name}"
                )
            node = TransformationChainResolver(
                self._stack
                + [TransformationChainResolver.Entry(name=base, value=child)]
            )
        return node if len(remainder) == 0 else node[remainder[0]]

    def resolve_depends_on(self) -> Optional[Union[sc.DataArray, sc.Variable]]:
        """
        Resolve the depends_on attribute of a transformation chain.

        Returns
        -------
        :
            The resolved position in meter, or None if no depends_on was found.
        """
        depends_on = self.value.get('depends_on')
        if depends_on is None:
            return None
        # Note that transformations have to be applied in "reverse" order, i.e.,
        # simply taking math.prod(chain) would be wrong, even if we could
        # ignore potential time-dependence.
        return combine_transformations(self.get_chain(depends_on))

    def get_chain(self, depends_on: str) -> List[Union[sc.DataArray, sc.Variable]]:
        if depends_on == '.':
            return []
        node = self[depends_on]
        transform = node.value.copy(deep=False)
        depends_on = '.'
        if isinstance(transform, sc.DataArray):
            if (attr := transform.coords.pop('depends_on', None)) is not None:
                depends_on = attr.value
            # If transform is time-dependent then we keep it is a DataArray, otherwise
            # we convert it to a Variable.
            transform = transform if transform.coords else transform.data
        if transform.dtype in (sc.DType.translation3, sc.DType.affine_transform3):
            transform = transform.to(unit='m', copy=False)
        return [transform] + node.parent.get_chain(depends_on)


def compute_positions(
    dg: sc.DataGroup,
    *,
    store_position: str = 'position',
    store_transform: Optional[str] = None,
) -> sc.DataGroup:
    """
    Recursively compute positions from depends_on attributes as well as the
    [xyz]_pixel_offset fields of NXdetector groups.

    This function does not operate directly on a NeXus file but on the result of
    loading a NeXus file or sub-group into a scipp.DataGroup. NeXus puts no
    limitations on the structure of the depends_on chains, i.e., they may reference
    parent groups. If this is the case, a call to this function will fail if only the
    subgroup is passed as input.

    Note that this does not consider "legacy" ways of storing positions. In particular,
    ``NXmonitor.distance``, ``NXdetector.distance``, ``NXdetector.polar_angle``, and
    ``NXdetector.azimuthal_angle`` are ignored.

    Note that transformation chains may be time-dependent. In this case it will not
    be applied to the pixel offsets, since the result may consume too much memory and
    the shape is in general incompatible with the shape of the data. Use the
    ``store_transform`` argument to store the resolved transformation chain in this
    case.

    If a transformation chain has an invalid 'depends_on' value, e.g., a path beyond
    the root data group, then the chain is ignored and no position is computed. This
    does not affect other chains.

    Parameters
    ----------
    dg:
        Data group with depends_on entry points into transformation chains.
    store_position:
        Name used to store result of resolving each depends_on chain.
    store_transform:
        If not None, store the resolved transformation chain in this field.

    Returns
    -------
    :
        New data group with added positions.
    """
    # Create resolver at root level, since any depends_on chain may lead to a parent,
    # i.e., we cannot use a resolver at the level of each chain's entry point.
    resolver = TransformationChainResolver.from_root(dg)
    return _with_positions(
        dg,
        store_position=store_position,
        store_transform=store_transform,
        resolver=resolver,
    )


def zip_pixel_offsets(x: Dict[str, sc.Variable], /) -> sc.Variable:
    """
    Zip the x_pixel_offset, y_pixel_offset, and z_pixel_offset fields into a vector.

    These fields originate from NXdetector groups. All but x_pixel_offset are optional,
    e.g., for 2D detectors. Zero values for missing fields are assumed.

    Parameters
    ----------
    mapping:
        Mapping (typically a data group, or data array coords) containing
        x_pixel_offset, y_pixel_offset, and z_pixel_offset.

    Returns
    -------
    :
        Vectors with pixel offsets.

    See Also
    --------
    compute_positions
    """
    zero = sc.scalar(0.0, unit=x['x_pixel_offset'].unit)
    return sc.spatial.as_vectors(
        x['x_pixel_offset'],
        x.get('y_pixel_offset', zero),
        x.get('z_pixel_offset', zero),
    )


def _with_positions(
    dg: sc.DataGroup,
    *,
    store_position: str,
    store_transform: Optional[str] = None,
    resolver: TransformationChainResolver,
) -> sc.DataGroup:
    out = sc.DataGroup()
    transform = None
    if 'depends_on' in dg:
        try:
            transform = resolver.resolve_depends_on()
            out[store_position] = transform * sc.vector([0, 0, 0], unit='m')
            if store_transform is not None:
                out[store_transform] = transform
        except TransformationChainResolver.ChainError as e:
            warnings.warn(
                UserWarning(f'depends_on chain references missing node:\n{e}'),
                stacklevel=2,
            )
    for name, value in dg.items():
        if isinstance(value, sc.DataGroup):
            value = _with_positions(
                value,
                store_position=store_position,
                store_transform=store_transform,
                resolver=resolver[name],
            )
        elif (
            isinstance(value, sc.DataArray)
            and 'x_pixel_offset' in value.coords
            # Transform can be time-dependent, do not apply it to offsets since
            # result can be massive and is in general not compatible with the shape
            # of the data.
            and (transform is not None and transform.dims == ())
        ):
            offset = zip_pixel_offsets(value.coords).to(unit='m', copy=False)
            value = value.assign_coords({store_position: transform * offset})
        out[name] = value
    return out



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Simon Heybrock
from __future__ import annotations

import datetime
import functools
import inspect
import re
import warnings
from typing import Any, Callable, Dict, List, Optional, Protocol, Tuple, Union, overload

import dateutil.parser
import h5py
import numpy as np
import scipp as sc

from .._common import convert_time_to_datetime64, to_plain_index
from .._hdf5_nexus import (
    _cset_to_encoding,
    _ensure_str,
    _ensure_supported_int_type,
    _warn_latin1_decode,
)
from ..typing import H5Dataset, H5Group, ScippIndex

NXobjectIndex = Union[str, ScippIndex]


def asarray(obj: Union[Any, sc.Variable]) -> sc.Variable:
    return obj if isinstance(obj, sc.Variable) else sc.scalar(obj, unit=None)


# TODO move into scipp
class DimensionedArray(Protocol):
    """
    A multi-dimensional array with a unit and dimension labels.

    Could be, e.g., a scipp.Variable or a dimple dataclass wrapping a numpy array.
    """

    @property
    def values(self):
        """Multi-dimensional array of values"""

    @property
    def unit(self):
        """Physical unit of the values"""

    @property
    def dims(self) -> List[str]:
        """Dimension labels for the values"""


class AttributeManager(Protocol):
    def __getitem__(self, name: str):
        """Get attribute"""


class NexusStructureError(Exception):
    """Invalid or unsupported class and field structure in Nexus."""

    pass


class Attrs:
    """HDF5 attributes."""

    def __init__(self, attrs: AttributeManager):
        self._attrs = attrs

    def __contains__(self, name: str) -> bool:
        return name in self._attrs

    def __getitem__(self, name: str) -> Any:
        attr = self._attrs[name]
        # Is this check for string attributes sufficient? Is there a better way?
        if isinstance(attr, (str, bytes)):
            cset = self._attrs.get_id(name.encode("utf-8")).get_type().get_cset()
            return _ensure_str(attr, _cset_to_encoding(cset))
        return attr

    def __setitem__(self, name: str, val: Any):
        self._attrs[name] = val

    def __iter__(self):
        yield from self._attrs

    def get(self, name: str, default=None) -> Any:
        return self[name] if name in self else default

    def keys(self):
        return self._attrs.keys()


def _is_time(obj):
    dummy = sc.empty(dims=[], shape=[], unit=obj.unit)
    try:
        dummy.to(unit='s')
        return True
    except sc.UnitError:
        return False


def _as_datetime(obj: Any):
    if isinstance(obj, str):
        try:
            # NumPy and scipp cannot handle timezone information. We therefore apply it,
            # i.e., convert to UTC.
            # Would like to use dateutil directly, but with Python's datetime we do not
            # get nanosecond precision. Therefore we combine numpy and dateutil parsing.
            date_only = 'T' not in obj
            if date_only:
                return sc.datetime(obj)
            date, time = obj.split('T')
            time_and_timezone_offset = re.split(r'Z|\+|-', time)
            time = time_and_timezone_offset[0]
            if len(time_and_timezone_offset) == 1:
                # No timezone, parse directly (scipp based on numpy)
                return sc.datetime(f'{date}T{time}')
            else:
                # There is timezone info. Parse with dateutil.
                dt = dateutil.parser.isoparse(obj)
                dt = dt.replace(microsecond=0)  # handled by numpy
                dt = dt.astimezone(datetime.timezone.utc)
                dt = dt.replace(tzinfo=None).isoformat()
                # We operate with string operations here and thus end up parsing date
                # and time twice. The reason is that the timezone-offset arithmetic
                # cannot be done, e.g., in nanoseconds without causing rounding errors.
                if '.' in time:
                    dt += f".{time.split('.')[1]}"
                return sc.datetime(dt)
        except ValueError:
            pass
    return None


def _dtype_from_dataset(dataset: H5Dataset) -> sc.DType:
    dtype = dataset.dtype
    if str(dtype).startswith('str') or h5py.check_string_dtype(dtype):
        dtype = sc.DType.string
    else:
        dtype = sc.DType(_ensure_supported_int_type(str(dtype)))
    return dtype


class Field:
    """NeXus field.

    In HDF5 fields are represented as dataset.
    """

    def __init__(
        self,
        dataset: H5Dataset,
        *,
        ancestor,
        dims=None,
        dtype: Optional[sc.DType] = None,
        is_time=None,
    ):
        self._ancestor = ancestor  # Usually the parent, but may be grandparent, etc.
        self._dataset = dataset
        self._dtype = _dtype_from_dataset(dataset) if dtype is None else dtype
        self._shape = self._dataset.shape
        if self._dtype == sc.DType.vector3:
            self._shape = self._shape[:-1]
        self._is_time = is_time
        # NeXus treats [] and [1] interchangeably. In general this is ill-defined, but
        # the best we can do appears to be squeezing unless the file provides names for
        # dimensions. The shape property of this class does thus not necessarily return
        # the same as the shape of the underlying dataset.
        if dims is not None:
            self._dims = tuple(dims)
            if len(self._dims) < len(self._shape):
                # The convention here is that the given dimensions apply to the shapes
                # starting from the left. So we only squeeze dimensions that are after
                # len(dims).
                self._shape = self._shape[: len(self._dims)] + tuple(
                    size for size in self._shape[len(self._dims) :] if size != 1
                )
        elif (axes := self.attrs.get('axes')) is not None:
            self._dims = tuple(axes.split(':'))
            # The standard says that the axes should be colon-separated, but some
            # files use comma-separated.
            if len(self._dims) == 1 and self._dataset.ndim > 1:
                self._dims = tuple(axes.split(','))
        else:
            self._shape = tuple(size for size in self._shape if size != 1)
            self._dims = tuple(f'dim_{i}' for i in range(self.ndim))

    def __getitem__(self, select) -> Union[Any, sc.Variable]:
        """Load the field as a :py:class:`scipp.Variable` or Python object.

        If the shape is empty and no unit is given this returns a Python object, such
        as a string or integer. Otherwise a :py:class:`scipp.Variable` is returned.
        """
        index = to_plain_index(self.dims, select)
        if isinstance(index, (int, slice)):
            index = (index,)

        base_dims = self.dims
        base_shape = self.shape
        dims = []
        shape = []
        for i, ind in enumerate(index):
            if not isinstance(ind, int):
                dims.append(base_dims[i])
                shape.append(len(range(*ind.indices(base_shape[i]))))

        variable = sc.empty(dims=dims, shape=shape, dtype=self.dtype, unit=self.unit)

        # If the variable is empty, return early
        if np.prod(shape) == 0:
            return variable

        if self.dtype == sc.DType.string:
            try:
                strings = self._dataset.asstr()[index]
            except UnicodeDecodeError as e:
                strings = self._dataset.asstr(encoding='latin-1')[index]
                _warn_latin1_decode(self._dataset, strings, str(e))
            variable.values = np.asarray(strings).flatten()
        elif variable.values.flags["C_CONTIGUOUS"]:
            # On versions of h5py prior to 3.2, a TypeError occurs in some cases
            # where h5py cannot broadcast data with e.g. shape (20, 1) to a buffer
            # of shape (20,). Note that broadcasting (1, 20) -> (20,) does work
            # (see https://github.com/h5py/h5py/pull/1796).
            # Therefore, we manually squeeze here.
            # A pin of h5py<3.2 is currently required by Mantid and hence scippneutron
            # (see https://github.com/h5py/h5py/issues/1880#issuecomment-823223154)
            # hence this workaround. Once we can use a more recent h5py with Mantid,
            # this try/except can be removed.
            try:
                self._dataset.read_direct(variable.values, source_sel=index)
            except TypeError:
                variable.values = self._dataset[index].squeeze()
        else:
            variable.values = self._dataset[index]
        if _is_time(variable):
            starts = []
            for name in self.attrs:
                if (dt := _as_datetime(self.attrs[name])) is not None:
                    starts.append(dt)
            if self._is_time and len(starts) == 0:
                starts.append(sc.epoch(unit=self.unit))
            if len(starts) == 1:
                variable = convert_time_to_datetime64(
                    variable,
                    start=starts[0],
                    scaling_factor=self.attrs.get('scaling_factor'),
                )
        if variable.ndim == 0 and variable.unit is None:
            # Work around scipp/scipp#2815, and avoid returning NumPy bool
            if isinstance(variable.values, np.ndarray) and variable.dtype != 'bool':
                return variable.values[()]
            else:
                return variable.value
        return variable

    def __repr__(self) -> str:
        return f'<Nexus field "{self._dataset.name}">'

    @property
    def attrs(self) -> Attrs:
        return Attrs(self._dataset.attrs)

    @property
    def dtype(self) -> sc.DType:
        return self._dtype

    @property
    def name(self) -> str:
        return self._dataset.name

    @property
    def file(self) -> NXroot:
        return NXroot(self._dataset.file)

    @property
    def parent(self) -> NXobject:
        return self._ancestor._make(self._dataset.parent)

    @property
    def ndim(self) -> int:
        """Total number of dimensions in the dataset.

        See the shape property for potential differences to the value returned by the
        underlying h5py.Dataset.ndim.
        """
        return len(self.shape)

    @property
    def shape(self) -> List[int]:
        """Shape of the field.

        NeXus may use extra dimensions of length one to store data, such as shape=[1]
        instead of shape=[]. This property returns the *squeezed* shape, dropping all
        length-1 dimensions that are not explicitly named. The returned shape may thus
        be different from the shape of the underlying h5py.Dataset.
        """
        return self._shape

    @property
    def dims(self) -> List[str]:
        return self._dims

    @property
    def sizes(self) -> Dict[str, int]:
        return dict(zip(self.dims, self.shape))

    @property
    def unit(self) -> Union[sc.Unit, None]:
        if (unit := self.attrs.get('units')) is not None:
            try:
                return sc.Unit(unit)
            except sc.UnitError:
                warnings.warn(
                    f"Unrecognized unit '{unit}' for value dataset "
                    f"in '{self.name}'; setting unit as 'dimensionless'",
                    stacklevel=2,
                )
                return sc.units.one
        return None


def is_dataset(obj: Union[H5Group, H5Dataset]) -> bool:
    """Return true if the object is an h5py.Dataset or equivalent.

    Use this instead of isinstance(obj, h5py.Dataset) to ensure that code is compatible
    with other h5py-alike interfaces.
    """
    return hasattr(obj, 'shape')


class NXobjectStrategy:
    @staticmethod
    def include_child(_) -> bool:
        """Return True if the child should be included when loading."""
        return True


class NXobject:
    """Base class for all NeXus groups."""

    def __init__(
        self,
        group: H5Group,
        *,
        definition: Any = None,
        strategy: Optional[Callable] = None,
    ):
        self._group = group
        # TODO can strategies replace child-params?
        self.child_params = {}
        self._definition = definition
        self._strategy = strategy
        if strategy is None and self._definition is not None:
            self._strategy = self._definition.make_strategy(self)
        if self._strategy is None:
            self._strategy = self._default_strategy()

    def _default_strategy(self):
        """
        Default strategy to use when none given and when the application definition
        does not provide one. Override in child classes to set a default.
        """
        return NXobjectStrategy

    def _make(self, group) -> NXobject:
        if (nx_class := Attrs(group.attrs).get('NX_class')) is not None:
            return _nx_class_registry().get(nx_class, NXobject)(
                group, definition=self._definition
            )
        return group  # Return underlying (h5py) group

    def _get_child(
        self, name: NXobjectIndex, use_field_dims: bool = False
    ) -> Union['NXobject', Field, sc.DataArray]:
        """Get item, with flag to control whether fields dims should be inferred"""
        if name is None:
            raise KeyError("None is not a valid index")
        if isinstance(name, str):
            item = self._group[name]
            if is_dataset(item):
                try:
                    dims = self._get_field_dims(name) if use_field_dims else None
                except Exception as e:
                    msg = (
                        f"Failed to determine axis names of {item.name}: {e}. "
                        "Falling back to default dimension labels."
                    )
                    warnings.warn(msg, stacklevel=2)
                    dims = None
                dtype = self._get_field_dtype(name)
                return Field(
                    item,
                    dims=dims,
                    dtype=dtype,
                    ancestor=self,
                    **self.child_params.get(name, {}),
                )
            else:
                return self._make(item)

        try:
            da = self._getitem(name)
            self._insert_leaf_properties(da)
        except NexusStructureError as e:
            # If the child class cannot load this group, we fall back to returning the
            # underlying datasets in a DataGroup.
            if type(self)._getitem == NXobject._getitem:
                raise
            else:
                msg = (
                    f"Failed to load {self.name} as {type(self).__name__}: {e} "
                    "Falling back to loading HDF5 group children as scipp.DataGroup."
                )
                warnings.warn(msg, stacklevel=2)
            da = NXobject._getitem(self, name)
        return da

    def _insert_leaf_properties(self, container):
        from .nexus_classes import NXgeometry
        from .nxcylindrical_geometry import NXcylindrical_geometry
        from .nxoff_geometry import NXoff_geometry

        def insert(container, name, obj):
            if hasattr(container, 'coords'):
                container.coords[name] = (
                    obj if isinstance(obj, sc.Variable) else sc.scalar(obj)
                )
            else:
                container[name] = obj

        detector_number = getattr(self, 'detector_number', None)
        if detector_number is not None:
            detector_number = container.coords[detector_number]
        for key, child in self[[NXcylindrical_geometry, NXoff_geometry]].items():
            insert(container, key, child.load_as_array(detector_number=detector_number))
        for key, child in self[NXgeometry].items():
            insert(container, key, child[()])
        if (t := self.depends_on) is not None:
            insert(container, 'depends_on', t)
            # If loading the transformation failed, 'depends_on' returns a string, the
            # path to the transformation. If this is a nested group, we load it here.
            # Note that this info is currently incomplete, since attributes are not
            # loaded.
            if isinstance(t, str):
                from .nexus_classes import NXtransformations

                for key, group in self[NXtransformations].items():
                    insert(container, key, group[()])

    def _get_children_by_nx_class(
        self, select: Union[type, List[type]]
    ) -> Dict[str, Union['NXobject', Field]]:
        children = {}
        select = tuple(select) if isinstance(select, list) else select
        for key in self.keys():
            if issubclass(type(self._get_child(key)), select):
                # Get child again via __getitem__ so correct field dims are used.
                children[key] = self[key]
        return children

    @overload
    def __getitem__(self, name: str) -> Union['NXobject', Field]:
        ...

    @overload
    def __getitem__(self, name: ScippIndex) -> Union[sc.DataArray, sc.DataGroup]:
        ...

    @overload
    def __getitem__(self, name: Union[type, List[type]]) -> Dict[str, 'NXobject']:
        ...

    def __getitem__(self, name):
        """
        Get a child group or child dataset, a selection of child groups, or load and
        return the current group.

        Three cases are supported:

        - String name: The child group or child dataset of that name is returned.
        - Class such as ``NXdata`` or ``NXlog``: A dict containing all direct children
          with a matching ``NX_class`` attribute are returned. Also accepts a tuple of
          classes. ``Field`` selects all child fields, i.e., all datasets but not
          groups.
        - Scipp-style index: Load the specified slice of the current group, returning
          a :class:`scipp.DataArray` or :class:`scipp.DataGroup`.

        Parameters
        ----------
        name:
            Child name, class, or index.

        Returns
        -------
        :
            Field, group, dict of fields, or loaded data.
        """

        def isclass(x):
            return inspect.isclass(x) and issubclass(x, (Field, NXobject))

        if isclass(name) or (
            isinstance(name, list) and len(name) and all(isclass(x) for x in name)
        ):
            return self._get_children_by_nx_class(name)
        return self._get_child(name, use_field_dims=True)

    def _getitem(self, index: ScippIndex) -> Union[sc.DataArray, sc.DataGroup]:
        include = getattr(self._strategy, 'include_child', lambda x: True)
        return sc.DataGroup(
            {name: child[index] for name, child in self.items() if include(child)}
        )

    def _get_field_dims(self, name: str) -> Union[None, List[str]]:
        """Subclasses should reimplement this to provide dimension labels for fields."""
        return None

    def _get_field_dtype(self, name: str) -> Union[None, sc.DType]:
        """Subclasses should reimplement this to override the dtype for fields."""
        return None

    def __contains__(self, name: str) -> bool:
        return name in self._group

    def get(self, name: str, default=None) -> Union['NXobject', Field, sc.DataArray]:
        return self[name] if name in self else default

    @property
    def attrs(self) -> Attrs:
        return Attrs(self._group.attrs)

    @property
    def name(self) -> str:
        return self._group.name

    @property
    def file(self) -> NXroot:
        return NXroot(self._group.file)

    @property
    def parent(self) -> NXobject:
        return self._make(self._group.parent)

    def _ipython_key_completions_(self) -> List[str]:
        return list(self.keys())

    def __iter__(self):
        yield from self._group

    def keys(self) -> List[str]:
        return self._group.keys()

    def values(self) -> List[Union[Field, 'NXobject']]:
        return [self[name] for name in self.keys()]

    def items(self) -> List[Tuple[str, Union[Field, 'NXobject']]]:
        return list(zip(self.keys(), self.values()))

    @property
    def nx_class(self) -> Optional[type]:
        """The value of the NX_class attribute of the group.

        In case of the subclass NXroot this returns 'NXroot' even if the attribute
        is not actually set. This is to support the majority of all legacy files, which
        do not have this attribute.
        """
        if (nxclass := self.attrs.get('NX_class')) is not None:
            return _nx_class_registry().get(nxclass)

    @property
    def depends_on(self) -> Union[sc.Variable, sc.DataArray, None]:
        if (depends_on := self.get('depends_on')) is not None:
            # Imported late to avoid cyclic import
            from .nxtransformations import TransformationError, get_full_transformation

            try:
                return get_full_transformation(depends_on)
            except (NexusStructureError, TransformationError) as e:
                warnings.warn(
                    f"Failed to load transformation {self.name}/{depends_on}:\n{e}\n"
                    "Falling back to returning the path to the transformation.",
                    stacklevel=2,
                )
                return depends_on[()]
        return None

    def __repr__(self) -> str:
        return f'<{type(self).__name__} "{self._group.name}">'

    def create_field(self, name: str, data: DimensionedArray, **kwargs) -> Field:
        if not isinstance(data, sc.Variable):
            return self._group.create_dataset(name, data=data, **kwargs)
        values = data.values
        if data.dtype == sc.DType.string:
            values = np.array(data.values, dtype=object)
        elif data.dtype == sc.DType.datetime64:
            start = sc.epoch(unit=data.unit)
            values = (data - start).values
        dataset = self._group.create_dataset(name, data=values, **kwargs)
        if data.unit is not None:
            dataset.attrs['units'] = str(data.unit)
        if data.dtype == sc.DType.datetime64:
            dataset.attrs['start'] = str(start.value)
        return Field(dataset, dims=data.dims, ancestor=self)

    def create_class(self, name: str, nx_class: Union[str, type]) -> NXobject:
        """Create empty HDF5 group with given name and set the NX_class attribute.

        Parameters
        ----------
        name:
            Group name.
        nx_class:
            Nexus class, can be a valid string for the NX_class attribute, or a
            subclass of NXobject, such as NXdata or NXlog.
        """
        group = self._group.create_group(name)
        attr = nx_class if isinstance(nx_class, str) else nx_class.__name__
        group.attrs['NX_class'] = attr
        return self._make(group)

    def __setitem__(self, name: str, value: Union[Field, NXobject, DimensionedArray]):
        """Create a link or a new field."""
        if isinstance(value, Field):
            self._group[name] = value._dataset
        elif isinstance(value, NXobject):
            self._group[name] = value._group
        elif hasattr(value, '__write_to_nexus_group__'):
            group = self.create_class(name, nx_class=value.nx_class)
            value.__write_to_nexus_group__(group)
        else:
            self.create_field(name, value)

    def __getattr__(self, attr: str) -> Union[Any, 'NXobject']:
        nxclass = _nx_class_registry().get(f'NX{attr}')
        if nxclass is None:
            raise AttributeError(f"'NXobject' object has no attribute {attr}")
        matches = self[nxclass]
        if len(matches) == 0:
            raise NexusStructureError(f"No group with requested NX_class='{nxclass}'")
        if len(matches) == 1:
            return next(iter(matches.values()))
        raise NexusStructureError(
            f"Multiple keys match {nxclass}, use obj[{nxclass}] "
            f"to obtain all matches instead of obj.{attr}."
        )

    def __dir__(self):
        keys = super().__dir__()
        nxclasses = []
        # Avoiding self.values() since it is more costly, but mainly since there may be
        # edge cases where creation of Field/NXobject may raise on unrelated children.
        for _, val in self._group.items():
            if not is_dataset(val):
                nxclasses.append(self._make(val).nx_class)
        for key in set(nxclasses):
            if key is None:
                continue
            if key in keys:
                continue
            if nxclasses.count(key) == 1:
                keys.append(key.__name__[2:])
        return keys


class NXroot(NXobject):
    """Root of a NeXus file."""

    @property
    def nx_class(self) -> type:
        # As an oversight in the NeXus standard and the reference implementation,
        # the NX_class was never set to NXroot. This applies to essentially all
        # files in existence before 2016, and files written by other implementations
        # that were inspired by the reference implementation. We thus hardcode NXroot:
        return NXroot


@functools.lru_cache()
def _nx_class_registry():
    from . import nexus_classes

    return dict(inspect.getmembers(nexus_classes, inspect.isclass))



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Simon Heybrock
from __future__ import annotations

import uuid
from itertools import chain
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union

import numpy as np
import scipp as sc

from ._cache import cached_property
from ._common import _to_canonical_select, convert_time_to_datetime64, to_child_select
from .base import (
    Group,
    NexusStructureError,
    NXobject,
    asvariable,
    base_definitions_dict,
)
from .field import Field, _is_time
from .nxevent_data import NXevent_data
from .typing import H5Dataset, ScippIndex


def _guess_dims(dims, shape, dataset: H5Dataset):
    """Guess dims of non-signal dataset based on shape."""
    if shape is None:
        return None
    if shape == dataset.shape:
        return dims
    if 1 < len(dataset.shape) <= len(shape):
        # We allow exact multi-dim matches at the end, even though there is a risk
        # for ambiguity with the transpose. Since image data is very common and many
        # sensors are square, this is a hopefully reasonable compromise.
        if dataset.shape == shape[-len(dataset.shape) :]:
            return dims[-len(dataset.shape) :]
    lut = {}
    for d, s in zip(dims, shape):
        if shape.count(s) == 1:
            lut[s] = d
    try:
        return [lut[s] for s in dataset.shape]
    except KeyError:
        try:  # Inner dimension may be bin-edges
            shape = list(dataset.shape)
            shape[-1] -= 1
            return [lut[s] for s in shape]
        except KeyError:
            pass
    return None


class NXdata(NXobject):
    def __init__(
        self,
        attrs: Dict[str, Any],
        children: Dict[str, Union[Field, Group]],
        fallback_dims: Optional[Tuple[str, ...]] = None,
        fallback_signal_name: Optional[str] = None,
    ):
        super().__init__(attrs=attrs, children=children)
        self._valid = True  # True if the children can be assembled
        self._signal_name = None
        self._signal = None
        self._aux_signals = list(attrs.get('auxiliary_signals', []))

        self._init_signal(
            name=attrs.get('signal', fallback_signal_name), children=children
        )
        if (errors := children.get('errors')) is not None:
            if (
                isinstance(self._signal, Field)
                and isinstance(errors, Field)
                and self._signal.errors is None
                and (errors.unit is None or self._signal.unit == errors.unit)
                and self._signal.dataset.shape == errors.dataset.shape
            ):
                self._signal.errors = errors.dataset
                del children['errors']
        self._init_axes(attrs=attrs, children=children)
        self._init_group_dims(attrs=attrs, fallback_dims=fallback_dims)

        for name, field in children.items():
            self._init_field_dims(name, field)

    def _init_field_dims(self, name: str, field: Union[Field, Group]) -> None:
        if not isinstance(field, Field):
            # If the NXdata contains subgroups we can generally not define valid
            # sizes... except for some non-signal "special fields" that return
            # a DataGroup that will be wrapped in a scalar Variable.
            if name == self._signal_name or name in self._aux_signals:
                return
            if field.attrs.get('NX_class') not in [
                'NXoff_geometry',
                'NXcylindrical_geometry',
                'NXgeometry',
                'NXtransformations',
            ]:
                self._valid = False
        elif (dims := self._get_dims(name, field)) is not None:
            # The convention here is that the given dimensions apply to the shapes
            # starting from the left. So we only squeeze dimensions that are after
            # len(dims).
            shape = _squeeze_trailing(dims, field.dataset.shape)
            field.sizes = dict(zip(dims, shape))
        elif self._valid:
            s1 = self._signal.sizes
            s2 = field.sizes
            if not set(s2.keys()).issubset(set(s1.keys())):
                self._valid = False
            elif any(s1[k] != s2[k] for k in s1.keys() & s2.keys()):
                self._valid = False

    def _init_signal(self, name: Optional[str], children):
        # There are multiple ways NeXus can define the "signal" dataset. The latest
        # version uses `signal` attribute on the group (passed as `name`). However,
        # we must give precedence to the `signal` attribute on the dataset, since
        # older files may use that (and the `signal` group attribute is unrelated).
        # Finally, NXlog and NXevent_data can take the role of the signal. In practice
        # those may not be indicate by a `signal` attribute, but we support that
        # anyway since otherwise we would not be able to find NXevent_data signals
        # in many common files.
        if name is not None and name in children:
            self._signal_name = name
            self._signal = children[name]
        # Legacy NXdata defines signal not as group attribute, but attr on dataset
        for name, field in children.items():
            # We ignore the signal value. Usually it is 1, but apparently one could
            # multiple signals. We do not support this, since it is legacy anyway.
            if 'signal' in field.attrs:
                self._signal_name = name
                self._signal = children[name]
                break
        # NXlog or NXevent_data can take the role of the signal.
        for name, field in children.items():
            if name == self._signal_name:
                # Avoid duplicate handling
                continue
            if isinstance(field, EventField) or (
                isinstance(field, Group) and field.nx_class in [NXlog, NXevent_data]
            ):
                if self._signal is None:
                    self._signal_name = name
                    self._signal = field
                else:
                    self._aux_signals.append(name)

    def _init_axes(
        self, attrs: Dict[str, Any], children: Dict[str, Union[Field, Group]]
    ):
        # Latest way of defining axes
        self._axes = attrs.get('axes')
        # Older way of defining axes
        self._signal_axes = (
            None if self._signal is None else self._signal.attrs.get('axes')
        )
        if self._signal_axes is not None:
            self._signal_axes = tuple(self._signal_axes.split(':'))
            # The standard says that the axes should be colon-separated, but some
            # files use comma-separated.
            if len(self._signal_axes) == 1 and self._signal.dataset.ndim > 1:
                self._signal_axes = tuple(self._signal_axes[0].split(','))
        # Another old way of defining axes. Apparently there are two different ways in
        # which this is used: A value of '1' indicates "this is an axis". As this would
        # not allow for determining an order, we have to assume that the signal field
        # has an "axes" attribute that defines the order. We can then ignore the "axis"
        # attributes, since they hold no further information. If there is not "axes"
        # attribute on the signal field then we have to assume that "axis" gives the
        # 1-based index of the axis.
        self._axis_index = {}
        if self._signal_axes is None:
            for name, field in children.items():
                if (axis := field.attrs.get('axis')) is not None:
                    self._axis_index[name] = axis

    def _get_named_axes(self, fallback_dims) -> Tuple[str, ...]:
        if self._axes is not None:
            # Unlike self.dims we *drop* entries that are '.'
            return tuple(a for a in self._axes if a != '.')
        elif self._signal_axes is not None:
            return self._signal_axes
        elif fallback_dims is not None:
            return fallback_dims
        else:
            return ()

    def _get_group_dims(self) -> Optional[Tuple[str, ...]]:
        """Try three ways of defining group dimensions."""
        # Apparently it is not possible to define dim labels unless there are
        # corresponding coords. Special case of '.' entries means "no coord".
        if self._axes is not None:
            return tuple(
                f'dim_{i}' if a == '.' else a for i, a in enumerate(self._axes)
            )
        if self._signal_axes is not None:
            return self._signal_axes
        if self._axis_index:
            return tuple(
                k for k, _ in sorted(self._axis_index.items(), key=lambda item: item[1])
            )
        return None

    def _init_group_dims(
        self, attrs: Dict[str, Any], fallback_dims: Optional[Tuple[str, ...]] = None
    ):
        group_dims = self._get_group_dims()

        if self._signal is None:
            self._valid = False
        elif isinstance(self._signal, EventField):
            group_dims = self._signal.dims
        else:
            if group_dims is not None:
                shape = self._signal.dataset.shape
                # If we have explicit group dims, we can drop trailing 1s.
                shape = _squeeze_trailing(group_dims, shape)
                self._signal.sizes = dict(zip(group_dims, shape))
            elif isinstance(self._signal, Group):
                group_dims = self._signal.dims
            elif fallback_dims is not None:
                shape = self._signal.dataset.shape
                group_dims = [
                    fallback_dims[i] if i < len(fallback_dims) else f'dim_{i}'
                    for i in range(len(shape))
                ]
                self._signal.sizes = dict(zip(group_dims, shape))

        self._group_dims = group_dims
        self._named_axes = self._get_named_axes(fallback_dims)

        indices_suffix = '_indices'
        indices_attrs = {
            key[: -len(indices_suffix)]: attr
            for key, attr in attrs.items()
            if key.endswith(indices_suffix)
        }
        # Datasets referenced by an "indices" attribute will be added as coords.
        self._explicit_coords = list(indices_attrs)

        dims = np.array(self._axes)
        self._dims_from_indices = {
            key: tuple(dims[np.array(indices).flatten()])
            for key, indices in indices_attrs.items()
        }

    def _get_dims(self, name, field):
        # Newly written files should always contain indices attributes, but the
        # standard recommends that readers should also make "best effort" guess
        # since legacy files do not set this attribute.
        if name == self._signal_name:
            return self._group_dims
        # Latest way of defining dims
        if (dims := self._dims_from_indices.get(name)) is not None:
            if '.' in dims:
                hdf5_dims = self._dims_from_hdf5(field)
                return tuple(
                    dim if dim != '.' else hdf5_dim
                    for dim, hdf5_dim in zip(dims, hdf5_dims)
                )
            return dims
        # Older way of defining dims via axis attribute
        if (axis := self._axis_index.get(name)) is not None:
            return (self._group_dims[axis - 1],)
        if name in self._aux_signals:
            return _guess_dims(
                self._group_dims, self._signal.dataset.shape, field.dataset
            )
        if name in self._named_axes:
            # If there are named axes then items of same name are "dimension
            # coordinates", i.e., have a dim matching their name.
            # However, if the item is not 1-D we need more labels. Try to use labels
            # of signal if dimensionality matches.
            if isinstance(self._signal, Field) and len(field.dataset.shape) == len(
                self._signal.dataset.shape
            ):
                return self._group_dims
            return (name,)
        if self._signal is not None and self._group_dims is not None:
            signal_shape = (
                self._signal.dataset.shape
                if isinstance(self._signal, Field)
                else (
                    self._signal.shape if isinstance(self._signal, EventField) else None
                )
            )
            return _guess_dims(self._group_dims, signal_shape, field.dataset)
        # While not mandated or recommended by the standard, we can try to find HDF5
        # dim labels as a fallback option for defining dimension labels. Ideally we
        # would like to do so in NXobject._init_field, but this causes significant
        # overhead for small files with many datasets. Defined here, this will only
        # take effect for NXdata, NXdetector, NXlog, and NXmonitor.
        return self._dims_from_hdf5(field)

    def _dims_from_hdf5(self, field):
        hdf5_dims = [dim.label for dim in getattr(field.dataset, 'dims', [])]
        if any(dim != '' for dim in hdf5_dims):
            while hdf5_dims and hdf5_dims[-1] == '':
                hdf5_dims.pop()
            return [f'dim_{i}' if dim == '' else dim for i, dim in enumerate(hdf5_dims)]

    @cached_property
    def sizes(self) -> Dict[str, int]:
        if not self._valid:
            return super().sizes
        sizes = dict(self._signal.sizes)
        for name in self._aux_signals:
            sizes.update(self._children[name].sizes)
        return sizes

    @property
    def unit(self) -> Union[None, sc.Unit]:
        return self._signal.unit if self._valid else super().unit

    def _bin_edge_dim(self, coord: Union[Any, Field]) -> Union[None, str]:
        if not isinstance(coord, Field):
            return None
        sizes = self.sizes
        for dim, size in zip(coord.dims, coord.shape):
            if (sz := sizes.get(dim)) is not None and sz + 1 == size:
                return dim
        return None

    def index_child(self, child: Union[Field, Group], sel: ScippIndex) -> ScippIndex:
        """Same as NXobject.index_child but also handles bin edges."""
        child_sel = to_child_select(
            tuple(self.sizes), child.dims, sel, bin_edge_dim=self._bin_edge_dim(child)
        )
        return child[child_sel]

    def assemble(
        self, dg: sc.DataGroup
    ) -> Union[sc.DataGroup, sc.DataArray, sc.Dataset]:
        return self._assemble_as_data(dg)

    def _assemble_as_data(
        self, dg: sc.DataGroup
    ) -> Union[sc.DataGroup, sc.DataArray, sc.Dataset]:
        if not self._valid:
            raise NexusStructureError("Could not determine signal field or dimensions.")
        dg = dg.copy(deep=False)
        aux = {name: dg.pop(name) for name in self._aux_signals}
        signal = dg.pop(self._signal_name)
        coords = dg
        if isinstance(signal, sc.DataGroup):
            raise NexusStructureError("Signal is not an array-like.")
        da = (
            signal
            if isinstance(signal, sc.DataArray)
            else sc.DataArray(data=asvariable(signal))
        )
        da = self._add_coords(da, coords)
        if aux:
            signals = {self._signal_name: da}
            signals.update(aux)
            if all(
                isinstance(v, (sc.Variable, sc.DataArray)) for v in signals.values()
            ):
                return sc.Dataset(signals)
            return sc.DataGroup(signals)
        return da

    def _assemble_as_physical_component(
        self, dg: sc.DataGroup, allow_in_coords: List[str]
    ) -> sc.DataGroup:
        """
        Assemble suitable for physical components such as NXdetector or NXmonitor.

        NeXus groups representing physical components typically contain fields or
        subgroups such as 'depends_on', 'NXtransformations', or 'NXoff_geometry'.
        Adding these wrapped into a scalar variable as coordinates would be
        cumbersome. Therefore, we do not assemble the entire NXdetector or NXmonitor
        into a single DataArray (as typically done by NXdata). On the other hand, we
        do want to return a useful DataArray with coordinates. In NeXus the data,
        coordinates, and other datasets or subgroups are all stored in the same group.
        As this is not very useful and user-friendly, we assemble the data and coords
        into a DataArray, which will take the place of the signal field in the output
        DataGroup, while non-signal non-coord fields are simply kept as-is.
        """
        data_items = sc.DataGroup()
        result = sc.DataGroup()
        forward = list(
            chain(
                [self._signal_name],
                self._group_dims or [],
                self._aux_signals,
                self._explicit_coords,
                allow_in_coords,
            )
        )

        for name, value in dg.items():
            if name in forward:
                data_items[name] = value
            else:
                result[name] = value
        assembled_nxdata = self._assemble_as_data(data_items)
        result.update(
            {self._signal_name: assembled_nxdata}
            if isinstance(assembled_nxdata, sc.DataArray)
            else assembled_nxdata
        )
        return result

    def _dim_of_coord(self, name: str, coord: sc.Variable) -> Union[None, str]:
        if len(coord.dims) == 1:
            return coord.dims[0]
        if name in coord.dims and name in self.dims:
            return name
        return self._bin_edge_dim(coord)

    def _should_be_aligned(
        self, da: sc.DataArray, name: str, coord: sc.Variable
    ) -> bool:
        if name == 'depends_on':
            return True
        dim_of_coord = self._dim_of_coord(name, coord)
        if dim_of_coord is None:
            return True
        if dim_of_coord not in da.dims:
            return False
        return True

    def _add_coords(self, da: sc.DataArray, coords: sc.DataGroup) -> sc.DataArray:
        """Add coords to a data array.

        Sets alignment in the same way as slicing scipp.DataArray would.
        """
        for name, coord in coords.items():
            if not isinstance(coord, sc.Variable):
                da.coords[name] = sc.scalar(coord)
            else:
                da.coords[name] = coord
                # We need the shape *before* slicing to determine dims, so we get the
                # field from the group for the conditional.
                da.coords.set_aligned(
                    name, self._should_be_aligned(da, name, self._children[name])
                )
        return da


def _squeeze_trailing(dims: Tuple[str, ...], shape: Tuple[int, ...]) -> Tuple[int, ...]:
    return shape[: len(dims)] + tuple(size for size in shape[len(dims) :] if size != 1)


class NXlog(NXdata):
    """
    NXlog, a time-series that can be loaded as a DataArray.

    In some cases the NXlog may contain additional time series, such as a connection
    status or alarm. These cannot be handled in a standard way, since the result cannot
    be represented as a single DataArray. Furthermore, they prevent positional
    time-indexing, since the time coord of each time-series is different. We can
    support label-based indexing for this in the future. If additional time-series
    are contained within the NXlog then loading will return a DataGroup of the
    individual time-series (DataArrays).
    """

    def __init__(self, attrs: Dict[str, Any], children: Dict[str, Union[Field, Group]]):
        children = dict(children)
        self._sublogs = []
        self._sublog_children = {}
        for name in children:
            if name.endswith('_time'):
                self._sublogs.append(name[:-5])
        # Extract all fields that belong to sublogs, since they will interfere with the
        # setup logic in the base class (NXdata).
        for name in self._sublogs:
            for k in list(children):
                if k.startswith(name):
                    field = children.pop(k)
                    self._init_field(field)
                    field.sizes = {
                        'time' if i == 0 else f'dim_{i}': size
                        for i, size in enumerate(field.dataset.shape)
                    }
                    self._sublog_children[k] = field

        super().__init__(
            attrs=attrs,
            children=children,
            fallback_dims=('time',),
            fallback_signal_name='value',
        )

    def read_children(self, sel: ScippIndex) -> sc.DataGroup:
        # Sublogs have distinct time axes (with a different length). Must disable
        # positional indexing.
        if self._sublogs and ('time' in _to_canonical_select(list(self.sizes), sel)):
            raise sc.DimensionError(
                "Cannot positionally select time since there are multiple "
                "time fields. Label-based selection is not supported yet."
            )
        dg = super().read_children(sel)
        for name, field in self._sublog_children.items():
            dg[name] = field[sel]
        return dg

    def _time_to_datetime(self, mapping):
        if (time := mapping.get('time')) is not None:
            if time.dtype != sc.DType.datetime64 and _is_time(time):
                mapping['time'] = convert_time_to_datetime64(
                    time, start=sc.epoch(unit=time.unit)
                )

    def _assemble_sublog(
        self, dg: sc.DataGroup, name: str, value_name: Optional[str] = None
    ) -> sc.DataArray:
        value_name = name if value_name is None else f'{name}_{value_name}'
        da = sc.DataArray(dg.pop(value_name), coords={'time': dg.pop(f'{name}_time')})
        for k in list(dg):
            if k.startswith(name):
                da.coords[k[len(name) + 1 :]] = dg.pop(k)
        self._time_to_datetime(da.coords)
        return da

    def assemble(
        self, dg: sc.DataGroup
    ) -> Union[sc.DataGroup, sc.DataArray, sc.Dataset]:
        self._time_to_datetime(dg)
        dg = sc.DataGroup(dg)
        sublogs = sc.DataGroup()
        for name in self._sublogs:
            # Somewhat arbitrary definition of which fields is the "value"
            value_name = 'severity' if name == 'alarm' else None
            sublogs[name] = self._assemble_sublog(dg, name, value_name=value_name)
        out = super().assemble(dg)
        return out if not sublogs else sc.DataGroup(value=out, **sublogs)


def _find_embedded_nxevent_data(
    children: Dict[str, Union[Field, Group]]
) -> Optional[Group]:
    if all(name in children for name in NXevent_data.mandatory_fields):
        parent = children['event_index'].parent._group
        event_group = Group(
            parent, definitions={'NXmonitor': NXevent_data, 'NXdetector': NXevent_data}
        )
        for name in list(children):
            if name in NXevent_data.handled_fields:
                del children[name]
        return event_group


class EventField:
    def __init__(self, event_data: Group, grouping_name: str, grouping: Field) -> None:
        """Create a field that represents an event data group.

        Parameters
        ----------
        event_data:
            The event data group holding the NXevent_data fields.
        grouping_name:
            The name of the field that contains the grouping information.
        grouping:
            The field that contains the grouping keys. These are IDs corresponding to
            the event_id field of the NXevent_data group, such as the detector_number
            field of an NXdetector.
        """
        self._event_data = event_data
        self._grouping_name = grouping_name
        self._grouping = grouping

    @property
    def attrs(self) -> Dict[str, Any]:
        return self._event_data.attrs

    @property
    def sizes(self) -> Dict[str, int]:
        return {**self._grouping.sizes, **self._event_data.sizes}

    @property
    def dims(self) -> Tuple[str, ...]:
        return self._grouping.dims + self._event_data.dims

    @property
    def shape(self) -> Tuple[int, ...]:
        return self._grouping.shape + self._event_data.shape

    def __getitem__(self, sel: ScippIndex) -> sc.DataArray:
        event_sel = to_child_select(self.dims, self._event_data.dims, sel)
        events = self._event_data[event_sel]
        detector_sel = to_child_select(self.dims, self._grouping.dims, sel)
        if not isinstance(events, sc.DataArray):
            return events
        da = _group_events(event_data=events, grouping=self._grouping[detector_sel])
        da.coords[self._grouping_name] = da.coords.pop('event_id')
        return da


class NXdetector(NXdata):
    _detector_number_fields = ['detector_number', 'pixel_id', 'spectrum_index']

    @staticmethod
    def _detector_number(children: Iterable[str]) -> Optional[str]:
        for name in NXdetector._detector_number_fields:
            if name in children:
                return name

    def __init__(self, attrs: Dict[str, Any], children: Dict[str, Union[Field, Group]]):
        fallback_dims = None
        if (det_num_name := NXdetector._detector_number(children)) is not None:
            if (detector_number := children[det_num_name]).dataset.ndim == 1:
                fallback_dims = (det_num_name,)
                detector_number.sizes = {det_num_name: detector_number.dataset.shape[0]}

        if (event_group := _find_embedded_nxevent_data(children)) is not None:
            embedded_events = uuid.uuid4().hex if 'events' in children else 'events'
            children[embedded_events] = event_group
        else:
            embedded_events = None

        def _maybe_event_field(name: str, child: Union[Field, Group]):
            if (
                name == embedded_events
                or (isinstance(child, Group) and child.nx_class == NXevent_data)
            ) and det_num_name is not None:
                event_field = EventField(
                    event_data=child,
                    grouping_name=det_num_name,
                    grouping=children.get(det_num_name),
                )
                return event_field
            return child

        children = {
            name: _maybe_event_field(name, child) for name, child in children.items()
        }

        super().__init__(
            attrs=attrs,
            children=children,
            fallback_dims=fallback_dims,
            fallback_signal_name='data',
        )

    def coord_allow_list(self) -> List[str]:
        """
        Names of datasets that will be treated as coordinates.

        Note that in addition to these, all datasets matching the data's dimensions
        as well as datasets explicitly referenced by an "indices" attribute in the
        group's list of attributes will be treated as coordinates.

        Override in subclasses to customize assembly of datasets into loaded output.
        """
        return NXdetector._detector_number_fields + [
            'time_of_flight',
            'raw_time_of_flight',
            'x_pixel_offset',
            'y_pixel_offset',
            'z_pixel_offset',
            'distance',
            'polar_angle',
            'azimuthal_angle',
            'crate',
            'slot',
            'input',
            'start_time',
            'stop_time',
            'sequence_number',
        ]

    def assemble(self, dg: sc.DataGroup) -> sc.DataGroup:
        bitmasks = {
            key[len('pixel_mask') :]: dg.pop(key)
            # tuple because we are going to change the dict over the iteration
            for key in tuple(dg)
            if key.startswith('pixel_mask')
        }

        out = self._assemble_as_physical_component(
            dg, allow_in_coords=self.coord_allow_list()
        )

        for suffix, bitmask in bitmasks.items():
            masks = self.transform_bitmask_to_dict_of_masks(bitmask, suffix)
            for signal in [self._signal_name] + self._aux_signals:
                for name, mask in masks.items():
                    out[signal].masks[name] = mask
        return out

    @staticmethod
    def transform_bitmask_to_dict_of_masks(bitmask: sc.Variable, suffix: str = ''):
        bit_to_mask_name = {
            0: 'gap_pixel',
            1: 'dead_pixel',
            2: 'underresponding_pixel',
            3: 'overresponding_pixel',
            4: 'noisy_pixel',
            6: 'part_of_a_cluster_of_problematic_pixels',
            8: 'user_defined_mask_pixel',
            31: 'virtual_pixel',
        }

        number_of_bits_in_dtype = 8 * bitmask.values.dtype.itemsize

        # Bitwise indicator of what masks are present
        masks_present = np.bitwise_or.reduce(bitmask.values.ravel())
        one = np.array(1)

        masks = {}
        for bit in range(number_of_bits_in_dtype):
            # Check if the mask associated with the current `bit` is present
            if masks_present & (one << bit):
                name = bit_to_mask_name.get(bit, f'undefined_bit{bit}_pixel') + suffix
                masks[name] = sc.array(
                    dims=bitmask.dims,
                    values=bitmask.values & (one << bit),
                    dtype='bool',
                )
        return masks

    @property
    def detector_number(self) -> Optional[str]:
        return self._detector_number(self._children)


class NXmonitor(NXdata):
    def __init__(self, attrs: Dict[str, Any], children: Dict[str, Union[Field, Group]]):
        if (event_group := _find_embedded_nxevent_data(children)) is not None:
            signal = uuid.uuid4().hex if 'events' in children else 'events'
            children[signal] = event_group
        else:
            signal = 'data'
        super().__init__(attrs=attrs, children=children, fallback_signal_name=signal)

    def coord_allow_list(self) -> List[str]:
        """
        Names of datasets that will be treated as coordinates.

        Note that in addition to these, all datasets matching the data's dimensions
        as well as datasets explicitly referenced by an "indices" attribute in the
        group's list of attributes will be treated as coordinates.

        Override in subclasses to customize assembly of datasets into loaded output.
        """
        return ['distance', 'time_of_flight']

    def assemble(self, dg: sc.DataGroup) -> sc.DataGroup:
        return self._assemble_as_physical_component(
            dg, allow_in_coords=self.coord_allow_list()
        )


def _group_events(
    *, event_data: sc.DataArray, grouping: Optional[sc.Variable] = None
) -> sc.DataArray:
    if grouping is None:
        event_id = 'event_id'
    else:
        # copy since sc.bin cannot deal with a non-contiguous view
        grouping = asvariable(grouping)
        event_id = grouping.flatten(to='event_id').copy()
    event_data.bins.coords['event_time_zero'] = sc.bins_like(
        event_data, fill_value=event_data.coords['event_time_zero']
    )
    # After loading raw NXevent_data it is guaranteed that the event table
    # is contiguous and that there is no masking. We can therefore use the
    # more efficient approach of binning from scratch instead of erasing the
    # 'event_time_zero' binning defined by NXevent_data.
    event_data = event_data.bins.constituents['data'].group(event_id)
    if grouping is None:
        return event_data
    return event_data.fold(dim='event_id', sizes=grouping.sizes)


base_definitions_dict['NXdata'] = NXdata
base_definitions_dict['NXlog'] = NXlog
base_definitions_dict['NXdetector'] = NXdetector
base_definitions_dict['NXmonitor'] = NXmonitor



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Simon Heybrock
from typing import Dict, Union

import scipp as sc
from scipp.spatial import linear_transform

from .leaf import Leaf
from .nxobject import ScippIndex

_matrix_units = dict(zip(['orientation_matrix', 'ub_matrix'], ['one', '1/Angstrom']))


class NXsample(Leaf):
    """Sample information, can be read as a dict."""

    def _getitem(
        self, select: ScippIndex
    ) -> Dict[str, Union[sc.Variable, sc.DataArray]]:
        content = super()._getitem(select)
        for key in _matrix_units:
            if (item := content.get(key)) is not None:
                content[key] = linear_transform(
                    value=item.values, unit=_matrix_units[key]
                )
        return content



import h5py
import numpy as np
import pytest
import scipp as sc
from scipp.testing import assert_identical

import scippnexus as snx
from scippnexus.nxtransformations import NXtransformations, TransformationChainResolver


def make_group(group: h5py.Group) -> snx.Group:
    return snx.Group(group, definitions=snx.base_definitions())


@pytest.fixture()
def h5root():
    """Yield h5py root group (file)"""
    with h5py.File('dummy.nxs', mode='w', driver="core", backing_store=False) as f:
        yield f


def create_detector(group):
    data = sc.array(dims=['xx', 'yy'], values=[[1.1, 2.2], [3.3, 4.4]])
    detector_numbers = sc.array(
        dims=['xx', 'yy'], unit=None, values=np.array([[1, 2], [3, 4]])
    )
    detector = snx.create_class(group, 'detector_0', snx.NXdetector)
    snx.create_field(detector, 'detector_number', detector_numbers)
    snx.create_field(detector, 'data', data)
    return detector


def test_Transformation_with_single_value(h5root):
    detector = create_detector(h5root)
    snx.create_field(
        detector, 'depends_on', sc.scalar('/detector_0/transformations/t1')
    )
    transformations = snx.create_class(detector, 'transformations', NXtransformations)
    value = sc.scalar(6.5, unit='mm')
    offset = sc.spatial.translation(value=[1, 2, 3], unit='mm')
    vector = sc.vector(value=[0, 0, 1])
    t = value * vector
    expected = sc.spatial.translations(dims=t.dims, values=t.values, unit=t.unit)
    expected = expected * offset
    value = snx.create_field(transformations, 't1', value)
    value.attrs['depends_on'] = '.'
    value.attrs['transformation_type'] = 'translation'
    value.attrs['offset'] = offset.values
    value.attrs['offset_units'] = str(offset.unit)
    value.attrs['vector'] = vector.value

    expected = sc.DataArray(data=expected, coords={'depends_on': sc.scalar('.')})
    detector = make_group(detector)
    depends_on = detector['depends_on'][()]
    assert depends_on == 'transformations/t1'
    t = detector[depends_on][()]
    assert_identical(t, expected)


def test_time_independent_Transformation_with_length_0(h5root):
    detector = create_detector(h5root)
    snx.create_field(
        detector, 'depends_on', sc.scalar('/detector_0/transformations/t1')
    )
    transformations = snx.create_class(detector, 'transformations', NXtransformations)
    value = sc.array(dims=['dim_0'], values=[], unit='mm')
    offset = sc.spatial.translation(value=[1, 2, 3], unit='mm')
    vector = sc.vector(value=[0, 0, 1])
    t = value * vector
    expected = sc.spatial.translations(dims=t.dims, values=t.values, unit=t.unit)
    expected = expected * offset
    value = snx.create_field(transformations, 't1', value)
    value.attrs['depends_on'] = '.'
    value.attrs['transformation_type'] = 'translation'
    value.attrs['offset'] = offset.values
    value.attrs['offset_units'] = str(offset.unit)
    value.attrs['vector'] = vector.value

    expected = sc.DataArray(data=expected, coords={'depends_on': sc.scalar('.')})
    detector = make_group(detector)
    depends_on = detector['depends_on'][()]
    assert depends_on == 'transformations/t1'
    t = detector[depends_on][()]
    assert_identical(t, expected)


def test_depends_on_absolute_path_to_sibling_group_resolved_to_relative_path(h5root):
    det1 = snx.create_class(h5root, 'det1', NXtransformations)
    snx.create_field(det1, 'depends_on', sc.scalar('/det2/transformations/t1'))

    depends_on = make_group(det1)['depends_on'][()]
    assert depends_on == '../det2/transformations/t1'


def test_depends_on_relative_path_unchanged(h5root):
    det1 = snx.create_class(h5root, 'det1', NXtransformations)
    snx.create_field(det1, 'depends_on', sc.scalar('transformations/t1'))

    depends_on = make_group(det1)['depends_on'][()]
    assert depends_on == 'transformations/t1'


def test_depends_on_attr_absolute_path_to_sibling_group_resolved_to_relative_path(
    h5root,
):
    det1 = snx.create_class(h5root, 'det1', NXtransformations)
    transformations = snx.create_class(det1, 'transformations', NXtransformations)
    t1 = snx.create_field(transformations, 't1', sc.scalar(0.1, unit='cm'))
    t1.attrs['depends_on'] = '/det2/transformations/t2'
    t1.attrs['transformation_type'] = 'translation'
    t1.attrs['vector'] = [0, 0, 1]

    loaded = make_group(det1)['transformations/t1'][()]
    assert loaded.coords['depends_on'].value == '../../det2/transformations/t2'


def test_depends_on_attr_relative_path_unchanged(h5root):
    det = snx.create_class(h5root, 'det', NXtransformations)
    transformations = snx.create_class(det, 'transformations', NXtransformations)
    t1 = snx.create_field(transformations, 't1', sc.scalar(0.1, unit='cm'))
    t1.attrs['depends_on'] = '.'
    t1.attrs['transformation_type'] = 'translation'
    t1.attrs['vector'] = [0, 0, 1]

    loaded = make_group(det)['transformations/t1'][()]
    assert loaded.coords['depends_on'].value == '.'
    t1.attrs['depends_on'] = 't2'
    loaded = make_group(det)['transformations/t1'][()]
    assert loaded.coords['depends_on'].value == 't2'


def test_chain_with_single_values_and_different_unit(h5root):
    detector = create_detector(h5root)
    snx.create_field(
        detector, 'depends_on', sc.scalar('/detector_0/transformations/t1')
    )
    transformations = snx.create_class(detector, 'transformations', NXtransformations)
    value = sc.scalar(6.5, unit='mm')
    offset = sc.spatial.translation(value=[1, 2, 3], unit='mm')
    vector = sc.vector(value=[0, 0, 1])
    t = value * vector
    value1 = snx.create_field(transformations, 't1', value)
    value1.attrs['depends_on'] = 't2'
    value1.attrs['transformation_type'] = 'translation'
    value1.attrs['offset'] = offset.values
    value1.attrs['offset_units'] = str(offset.unit)
    value1.attrs['vector'] = vector.value
    value2 = snx.create_field(transformations, 't2', value.to(unit='cm'))
    value2.attrs['depends_on'] = '.'
    value2.attrs['transformation_type'] = 'translation'
    value2.attrs['vector'] = vector.value

    t1 = sc.spatial.translations(dims=t.dims, values=t.values, unit=t.unit) * offset
    t2 = sc.spatial.translations(dims=t.dims, values=t.values, unit=t.unit).to(
        unit='cm'
    )
    detector = make_group(h5root['detector_0'])
    loaded = detector[()]
    depends_on = loaded['depends_on']
    assert depends_on == 'transformations/t1'
    transforms = loaded['transformations']
    assert_identical(transforms['t1'].data, t1)
    assert transforms['t1'].coords['depends_on'].value == 't2'
    assert_identical(transforms['t2'].data, t2)
    assert transforms['t2'].coords['depends_on'].value == '.'


def test_Transformation_with_multiple_values(h5root):
    detector = create_detector(h5root)
    snx.create_field(
        detector, 'depends_on', sc.scalar('/detector_0/transformations/t1')
    )
    transformations = snx.create_class(detector, 'transformations', NXtransformations)
    log = sc.DataArray(
        sc.array(dims=['time'], values=[1.1, 2.2], unit='m'),
        coords={'time': sc.array(dims=['time'], values=[11, 22], unit='s')},
    )
    log.coords['time'] = sc.epoch(unit='ns') + log.coords['time'].to(unit='ns')
    offset = sc.spatial.translation(value=[1, 2, 3], unit='m')
    vector = sc.vector(value=[0, 0, 1])
    t = log * vector
    t.data = sc.spatial.translations(dims=t.dims, values=t.values, unit=t.unit)
    value = snx.create_class(transformations, 't1', snx.NXlog)
    snx.create_field(value, 'time', log.coords['time'] - sc.epoch(unit='ns'))
    snx.create_field(value, 'value', log.data)
    value.attrs['depends_on'] = '.'
    value.attrs['transformation_type'] = 'translation'
    value.attrs['offset'] = offset.values
    value.attrs['offset_units'] = str(offset.unit)
    value.attrs['vector'] = vector.value

    expected = t * offset
    expected.coords['depends_on'] = sc.scalar('.')
    detector = make_group(detector)
    depends_on = detector['depends_on'][()]
    assert depends_on == 'transformations/t1'
    assert_identical(detector[depends_on][()], expected)


def test_chain_with_multiple_values(h5root):
    detector = create_detector(h5root)
    snx.create_field(
        detector, 'depends_on', sc.scalar('/detector_0/transformations/t1')
    )
    transformations = snx.create_class(detector, 'transformations', NXtransformations)
    log = sc.DataArray(
        sc.array(dims=['time'], values=[1.1, 2.2], unit='m'),
        coords={'time': sc.array(dims=['time'], values=[11, 22], unit='s')},
    )
    log.coords['time'] = sc.epoch(unit='ns') + log.coords['time'].to(unit='ns')
    offset = sc.spatial.translation(value=[1, 2, 3], unit='m')
    vector = sc.vector(value=[0, 0, 1])
    t = log * vector
    t.data = sc.spatial.translations(dims=t.dims, values=t.values, unit=t.unit)
    value1 = snx.create_class(transformations, 't1', snx.NXlog)
    snx.create_field(value1, 'time', log.coords['time'] - sc.epoch(unit='ns'))
    snx.create_field(value1, 'value', log.data)
    value1.attrs['depends_on'] = 't2'
    value1.attrs['transformation_type'] = 'translation'
    value1.attrs['offset'] = offset.values
    value1.attrs['offset_units'] = str(offset.unit)
    value1.attrs['vector'] = vector.value
    value2 = snx.create_class(transformations, 't2', snx.NXlog)
    snx.create_field(value2, 'time', log.coords['time'] - sc.epoch(unit='ns'))
    snx.create_field(value2, 'value', log.data)
    value2.attrs['depends_on'] = '.'
    value2.attrs['transformation_type'] = 'translation'
    value2.attrs['vector'] = vector.value

    expected1 = t * offset
    expected1.coords['depends_on'] = sc.scalar('t2')
    expected2 = t
    expected2.coords['depends_on'] = sc.scalar('.')
    detector = make_group(detector)[()]
    depends_on = detector['depends_on']
    assert depends_on == 'transformations/t1'
    assert_identical(detector['transformations']['t1'], expected1)
    assert_identical(detector['transformations']['t2'], expected2)


def test_chain_with_multiple_values_and_different_time_unit(h5root):
    detector = create_detector(h5root)
    snx.create_field(
        detector, 'depends_on', sc.scalar('/detector_0/transformations/t1')
    )
    transformations = snx.create_class(detector, 'transformations', NXtransformations)
    # Making sure to not use nanoseconds since that is used internally and may thus
    # mask bugs.
    log = sc.DataArray(
        sc.array(dims=['time'], values=[1.1, 2.2], unit='m'),
        coords={'time': sc.array(dims=['time'], values=[11, 22], unit='s')},
    )
    log.coords['time'] = sc.epoch(unit='us') + log.coords['time'].to(unit='us')
    offset = sc.spatial.translation(value=[1, 2, 3], unit='m')
    vector = sc.vector(value=[0, 0, 1])
    t = log * vector
    t.data = sc.spatial.translations(dims=t.dims, values=t.values, unit=t.unit)
    value1 = snx.create_class(transformations, 't1', snx.NXlog)
    snx.create_field(value1, 'time', log.coords['time'] - sc.epoch(unit='us'))
    snx.create_field(value1, 'value', log.data)
    value1.attrs['depends_on'] = 't2'
    value1.attrs['transformation_type'] = 'translation'
    value1.attrs['offset'] = offset.values
    value1.attrs['offset_units'] = str(offset.unit)
    value1.attrs['vector'] = vector.value
    value2 = snx.create_class(transformations, 't2', snx.NXlog)
    snx.create_field(
        value2, 'time', log.coords['time'].to(unit='ms') - sc.epoch(unit='ms')
    )
    snx.create_field(value2, 'value', log.data)
    value2.attrs['depends_on'] = '.'
    value2.attrs['transformation_type'] = 'translation'
    value2.attrs['vector'] = vector.value

    expected1 = t * offset
    expected1.coords['depends_on'] = sc.scalar('t2')

    t2 = t.copy()
    t2.coords['time'] = t2.coords['time'].to(unit='ms')
    expected2 = t2
    expected2.coords['depends_on'] = sc.scalar('.')

    detector = make_group(detector)
    loaded = detector[...]
    depends_on = loaded['depends_on']
    assert depends_on == 'transformations/t1'
    assert_identical(loaded['transformations']['t1'], expected1)
    assert_identical(loaded['transformations']['t2'], expected2)


def test_broken_time_dependent_transformation_returns_datagroup_but_sets_up_depends_on(
    h5root,
):
    detector = create_detector(h5root)
    snx.create_field(
        detector, 'depends_on', sc.scalar('/detector_0/transformations/t1')
    )
    transformations = snx.create_class(detector, 'transformations', NXtransformations)
    log = sc.DataArray(
        sc.array(dims=['time'], values=[1.1, 2.2], unit='m'),
        coords={'time': sc.array(dims=['time'], values=[11, 22], unit='s')},
    )
    log.coords['time'] = sc.epoch(unit='ns') + log.coords['time'].to(unit='ns')
    offset = sc.spatial.translation(value=[1, 2, 3], unit='m')
    vector = sc.vector(value=[0, 0, 1])
    t = log * vector
    t.data = sc.spatial.translations(dims=t.dims, values=t.values, unit=t.unit)
    value = snx.create_class(transformations, 't1', snx.NXlog)
    snx.create_field(value, 'time', log.coords['time'] - sc.epoch(unit='ns'))
    # This makes the transform "broken" since "time" has length 2 but data has length 0.
    snx.create_field(value, 'value', log.data[0:0])
    value.attrs['depends_on'] = '.'
    value.attrs['transformation_type'] = 'translation'
    value.attrs['offset'] = offset.values
    value.attrs['offset_units'] = str(offset.unit)
    value.attrs['vector'] = vector.value

    detector = make_group(detector)
    loaded = detector[()]
    t = loaded['transformations']
    assert isinstance(t, sc.DataGroup)
    # Due to the way NXtransformations works, vital information is stored in the
    # attributes. DataGroup does currently not support attributes, so this information
    # is mostly useless until that is addressed.
    t1 = t['t1']
    assert isinstance(t1, sc.DataGroup)
    assert t1.keys() == {'time', 'value'}
    assert loaded['depends_on'] == 'transformations/t1'
    assert_identical(loaded['transformations']['t1'], t1)


def write_translation(
    group, name: str, value: sc.Variable, offset: sc.Variable, vector: sc.Variable
) -> None:
    dset = snx.create_field(group, name, value)
    dset.attrs['transformation_type'] = 'translation'
    dset.attrs['offset'] = offset.values
    dset.attrs['offset_units'] = str(offset.unit)
    dset.attrs['vector'] = vector.value


def test_nxtransformations_group_single_item(h5root):
    value = sc.scalar(2.4, unit='mm')
    offset = sc.spatial.translation(value=[6, 2, 6], unit='mm')
    vector = sc.vector(value=[0, 1, 1])
    t = value * vector
    expected = (
        sc.spatial.translations(dims=t.dims, values=t.values, unit=t.unit) * offset
    )

    transformations = snx.create_class(h5root, 'transformations', NXtransformations)
    write_translation(transformations, 't1', value, offset, vector)

    loaded = make_group(h5root)['transformations'][()]
    assert set(loaded.keys()) == {'t1'}
    assert sc.identical(loaded['t1'], expected)


def test_nxtransformations_group_two_independent_items(h5root):
    transformations = snx.create_class(h5root, 'transformations', NXtransformations)

    value = sc.scalar(2.4, unit='mm')
    offset = sc.spatial.translation(value=[6, 2, 6], unit='mm')
    vector = sc.vector(value=[0, 1, 1])
    t = value * vector
    write_translation(transformations, 't1', value, offset, vector)
    expected1 = (
        sc.spatial.translations(dims=t.dims, values=t.values, unit=t.unit) * offset
    )

    value = value * 0.1
    t = value * vector
    write_translation(transformations, 't2', value, offset, vector)
    expected2 = (
        sc.spatial.translations(dims=t.dims, values=t.values, unit=t.unit) * offset
    )

    loaded = make_group(h5root)['transformations'][()]
    assert set(loaded.keys()) == {'t1', 't2'}
    assert sc.identical(loaded['t1'], expected1)
    assert sc.identical(loaded['t2'], expected2)


def test_nxtransformations_group_single_chain(h5root):
    transformations = snx.create_class(h5root, 'transformations', NXtransformations)

    value = sc.scalar(2.4, unit='mm')
    offset = sc.spatial.translation(value=[6, 2, 6], unit='mm')
    vector = sc.vector(value=[0, 1, 1])
    t = value * vector
    write_translation(transformations, 't1', value, offset, vector)
    expected1 = (
        sc.spatial.translations(dims=t.dims, values=t.values, unit=t.unit) * offset
    )

    value = value * 0.1
    t = value * vector
    write_translation(transformations, 't2', value, offset, vector)
    transformations['t2'].attrs['depends_on'] = 't1'
    expected2 = (
        sc.spatial.translations(dims=t.dims, values=t.values, unit=t.unit) * offset
    )

    loaded = make_group(h5root)['transformations'][()]
    assert set(loaded.keys()) == {'t1', 't2'}
    assert_identical(loaded['t1'], expected1)
    assert_identical(loaded['t2'].data, expected2)
    assert loaded['t2'].coords['depends_on'].value == 't1'


def test_slice_transformations(h5root):
    transformations = snx.create_class(h5root, 'transformations', NXtransformations)
    log = sc.DataArray(
        sc.array(dims=['time'], values=[1.1, 2.2, 3.3], unit='m'),
        coords={'time': sc.array(dims=['time'], values=[11, 22, 33], unit='s')},
    )
    log.coords['time'] = sc.epoch(unit='ns') + log.coords['time'].to(unit='ns')
    offset = sc.spatial.translation(value=[1, 2, 3], unit='m')
    vector = sc.vector(value=[0, 0, 1])
    t = log * vector
    t.data = sc.spatial.translations(dims=t.dims, values=t.values, unit=t.unit)
    value1 = snx.create_class(transformations, 't1', snx.NXlog)
    snx.create_field(value1, 'time', log.coords['time'] - sc.epoch(unit='ns'))
    snx.create_field(value1, 'value', log.data)
    value1.attrs['transformation_type'] = 'translation'
    value1.attrs['offset'] = offset.values
    value1.attrs['offset_units'] = str(offset.unit)
    value1.attrs['vector'] = vector.value

    expected = t * offset

    assert sc.identical(
        make_group(h5root)['transformations']['time', 1:3]['t1'], expected['time', 1:3]
    )


def test_TransformationChainResolver_path_handling():
    tree = TransformationChainResolver.from_root({'a': {'b': {'c': 1}}})
    assert tree['a']['b']['c'].value == 1
    assert tree['a/b/c'].value == 1
    assert tree['/a/b/c'].value == 1
    assert tree['a']['../a/b/c'].value == 1
    assert tree['a/b']['../../a/b/c'].value == 1
    assert tree['a/b']['./c'].value == 1


def test_TransformationChainResolver_name():
    tree = TransformationChainResolver.from_root({'a': {'b': {'c': 1}}})
    assert tree['a']['b']['c'].name == '/a/b/c'
    assert tree['a/b/c'].name == '/a/b/c'
    assert tree['/a/b/c'].name == '/a/b/c'
    assert tree['a']['../a/b/c'].name == '/a/b/c'
    assert tree['a/b']['../../a/b/c'].name == '/a/b/c'
    assert tree['a/b']['./c'].name == '/a/b/c'


def test_TransformationChainResolver_raises_ChainError_if_child_does_not_exists():
    tree = TransformationChainResolver.from_root({'a': {'b': {'c': 1}}})
    with pytest.raises(TransformationChainResolver.ChainError):
        tree['a']['b']['d']


def test_TransformationChainResolver_raises_ChainError_if_path_leads_beyond_root():
    tree = TransformationChainResolver.from_root({'a': {'b': {'c': 1}}})
    with pytest.raises(TransformationChainResolver.ChainError):
        tree['..']
    with pytest.raises(TransformationChainResolver.ChainError):
        tree['a']['../..']
    with pytest.raises(TransformationChainResolver.ChainError):
        tree['../a']


origin = sc.vector([0, 0, 0], unit='m')
shiftX = sc.spatial.translation(value=[1, 0, 0], unit='m')
rotZ = sc.spatial.rotations_from_rotvecs(sc.vector([0, 0, 90], unit='deg'))


def test_resolve_depends_on_dot():
    tree = TransformationChainResolver.from_root({'depends_on': '.'})
    assert sc.identical(tree.resolve_depends_on() * origin, origin)


def test_resolve_depends_on_child():
    transform = sc.DataArray(shiftX, coords={'depends_on': sc.scalar('.')})
    tree = TransformationChainResolver.from_root(
        {'depends_on': 'child', 'child': transform}
    )
    expected = sc.vector([1, 0, 0], unit='m')
    assert sc.identical(tree.resolve_depends_on() * origin, expected)


def test_resolve_depends_on_grandchild():
    transform = sc.DataArray(shiftX, coords={'depends_on': sc.scalar('.')})
    tree = TransformationChainResolver.from_root(
        {'depends_on': 'child/grandchild', 'child': {'grandchild': transform}}
    )
    expected = sc.vector([1, 0, 0], unit='m')
    assert sc.identical(tree.resolve_depends_on() * origin, expected)


def test_resolve_depends_on_child1_depends_on_child2():
    transform1 = sc.DataArray(shiftX, coords={'depends_on': sc.scalar('child2')})
    transform2 = sc.DataArray(rotZ, coords={'depends_on': sc.scalar('.')})
    tree = TransformationChainResolver.from_root(
        {'depends_on': 'child1', 'child1': transform1, 'child2': transform2}
    )
    # Note order
    expected = transform2.data * transform1.data
    assert sc.identical(tree.resolve_depends_on(), expected)


def test_resolve_depends_on_grandchild1_depends_on_grandchild2():
    transform1 = sc.DataArray(shiftX, coords={'depends_on': sc.scalar('grandchild2')})
    transform2 = sc.DataArray(rotZ, coords={'depends_on': sc.scalar('.')})
    tree = TransformationChainResolver.from_root(
        {
            'depends_on': 'child/grandchild1',
            'child': {'grandchild1': transform1, 'grandchild2': transform2},
        }
    )
    expected = transform2.data * transform1.data
    assert sc.identical(tree.resolve_depends_on(), expected)


def test_resolve_depends_on_grandchild1_depends_on_child2():
    transform1 = sc.DataArray(shiftX, coords={'depends_on': sc.scalar('../child2')})
    transform2 = sc.DataArray(rotZ, coords={'depends_on': sc.scalar('.')})
    tree = TransformationChainResolver.from_root(
        {
            'depends_on': 'child1/grandchild1',
            'child1': {'grandchild1': transform1},
            'child2': transform2,
        }
    )
    expected = transform2.data * transform1.data
    assert sc.identical(tree.resolve_depends_on(), expected)


def test_compute_positions(h5root):
    instrument = snx.create_class(h5root, 'instrument', snx.NXinstrument)
    detector = create_detector(instrument)
    snx.create_field(detector, 'x_pixel_offset', sc.linspace('xx', -1, 1, 2, unit='m'))
    snx.create_field(detector, 'y_pixel_offset', sc.linspace('yy', -1, 1, 2, unit='m'))
    detector.attrs['axes'] = ['xx', 'yy']
    detector.attrs['x_pixel_offset_indices'] = [0]
    detector.attrs['y_pixel_offset_indices'] = [1]
    snx.create_field(
        detector, 'depends_on', sc.scalar('/instrument/detector_0/transformations/t1')
    )
    transformations = snx.create_class(detector, 'transformations', NXtransformations)
    value = sc.scalar(6.5, unit='mm')
    offset = sc.spatial.translation(value=[1, 2, 3], unit='mm')
    vector = sc.vector(value=[0, 0, 1])
    t = value * vector
    value1 = snx.create_field(transformations, 't1', value)
    value1.attrs['depends_on'] = 't2'
    value1.attrs['transformation_type'] = 'translation'
    value1.attrs['offset'] = offset.values
    value1.attrs['offset_units'] = str(offset.unit)
    value1.attrs['vector'] = vector.value
    value2 = snx.create_field(transformations, 't2', value.to(unit='cm'))
    value2.attrs['depends_on'] = '.'
    value2.attrs['transformation_type'] = 'translation'
    value2.attrs['vector'] = vector.value

    t1 = sc.spatial.translations(dims=t.dims, values=t.values, unit=t.unit) * offset
    t2 = sc.spatial.translations(dims=t.dims, values=t.values, unit=t.unit).to(
        unit='cm'
    )
    root = make_group(h5root)
    loaded = root[()]
    result = snx.compute_positions(loaded)
    origin = sc.vector([0, 0, 0], unit='m')
    assert_identical(
        result['instrument']['detector_0']['position'],
        t2.to(unit='m') * t1.to(unit='m') * origin,
    )
    assert_identical(
        result['instrument']['detector_0']['data'].coords['position'],
        t2.to(unit='m')
        * t1.to(unit='m')
        * sc.vectors(
            dims=['xx', 'yy'],
            values=[[[-1, -1, 0], [-1, 1, 0]], [[1, -1, 0], [1, 1, 0]]],
            unit='m',
        ),
    )


def test_compute_positions_with_rotation(h5root):
    instrument = snx.create_class(h5root, 'instrument', snx.NXinstrument)
    detector = create_detector(instrument)
    snx.create_field(detector, 'x_pixel_offset', sc.linspace('xx', -1, 1, 2, unit='m'))
    snx.create_field(detector, 'y_pixel_offset', sc.linspace('yy', -1, 1, 2, unit='m'))
    detector.attrs['axes'] = ['xx', 'yy']
    detector.attrs['x_pixel_offset_indices'] = [0]
    detector.attrs['y_pixel_offset_indices'] = [1]
    snx.create_field(
        detector, 'depends_on', sc.scalar('/instrument/detector_0/transformations/t1')
    )
    transformations = snx.create_class(detector, 'transformations', NXtransformations)
    value = sc.scalar(90.0, unit='deg')
    vector = sc.vector(value=[0, 1, 0])
    rot = snx.create_field(transformations, 't1', value)
    rot.attrs['depends_on'] = '.'
    rot.attrs['transformation_type'] = 'rotation'
    rot.attrs['vector'] = vector.value

    transform = sc.spatial.rotations_from_rotvecs(sc.vector([0, 90, 0], unit='deg'))
    root = make_group(h5root)
    loaded = root[()]
    result = snx.compute_positions(loaded)
    origin = sc.vector([0, 0, 0], unit='m')
    assert_identical(result['instrument']['detector_0']['position'], origin)
    assert_identical(
        result['instrument']['detector_0']['data'].coords['position'],
        transform
        * sc.vectors(
            dims=['xx', 'yy'],
            values=[[[-1, -1, 0], [-1, 1, 0]], [[1, -1, 0], [1, 1, 0]]],
            unit='m',
        ),
    )


@pytest.mark.filterwarnings("ignore:depends_on chain references missing node")
def test_compute_positions_skips_for_path_beyond_root(h5root):
    instrument = snx.create_class(h5root, 'instrument', snx.NXinstrument)
    value = sc.scalar(6.5, unit='m')
    vector = sc.vector(value=[0, 0, 1])
    transform1 = snx.create_field(h5root, 't1', value)
    transform1.attrs['depends_on'] = '.'
    transform1.attrs['transformation_type'] = 'translation'
    transform1.attrs['vector'] = vector.value
    transform2 = snx.create_field(instrument, 't2', value)
    transform2.attrs['depends_on'] = '.'
    transform2.attrs['transformation_type'] = 'translation'
    transform2.attrs['vector'] = vector.value
    monitor1 = snx.create_class(instrument, 'monitor1', snx.NXmonitor)
    monitor2 = snx.create_class(instrument, 'monitor2', snx.NXmonitor)
    snx.create_field(monitor1, 'depends_on', '../../t1')
    snx.create_field(monitor2, 'depends_on', '../t2')
    root = make_group(h5root)
    loaded = root[()]
    assert 'position' in snx.compute_positions(loaded)['instrument']['monitor1']
    assert 'position' in snx.compute_positions(loaded)['instrument']['monitor2']
    assert 'position' not in snx.compute_positions(loaded['instrument'])['monitor1']
    assert 'position' in snx.compute_positions(loaded['instrument'])['monitor2']


def test_compute_positions_returns_position_with_unit_meters(h5root):
    instrument = snx.create_class(h5root, 'instrument', snx.NXinstrument)
    value = sc.scalar(6.5, unit='cm')
    vector = sc.vector(value=[0, 0, 1])
    transform = snx.create_field(instrument, 't1', value)
    transform.attrs['depends_on'] = '.'
    transform.attrs['transformation_type'] = 'translation'
    transform.attrs['vector'] = vector.value
    monitor = snx.create_class(instrument, 'monitor', snx.NXmonitor)
    snx.create_field(monitor, 'depends_on', '../t1')
    root = make_group(h5root)
    loaded = root[()]
    mon = snx.compute_positions(loaded)['instrument']['monitor']
    assert mon['position'].unit == 'm'


def test_compute_positions_handles_chains_with_mixed_units(h5root):
    instrument = snx.create_class(h5root, 'instrument', snx.NXinstrument)
    vector = sc.vector(value=[0, 0, 1])
    t1 = snx.create_field(instrument, 't1', sc.scalar(100, unit='cm'))
    t1.attrs['depends_on'] = 't2'
    t1.attrs['transformation_type'] = 'translation'
    t1.attrs['vector'] = vector.value
    t2 = snx.create_field(instrument, 't2', sc.scalar(1000, unit='mm'))
    t2.attrs['depends_on'] = '.'
    t2.attrs['transformation_type'] = 'translation'
    t2.attrs['vector'] = vector.value
    monitor = snx.create_class(instrument, 'monitor', snx.NXmonitor)
    snx.create_field(monitor, 'depends_on', '../t1')
    root = make_group(h5root)
    loaded = root[()]
    mon = snx.compute_positions(loaded)['instrument']['monitor']
    assert_identical(mon['position'], sc.vector([0, 0, 2], unit='m'))


def test_compute_positions_does_not_apply_time_dependent_transform_to_pixel_offsets(
    h5root,
):
    detector = create_detector(h5root)
    snx.create_field(detector, 'x_pixel_offset', sc.linspace('xx', -1, 1, 2, unit='m'))
    snx.create_field(detector, 'y_pixel_offset', sc.linspace('yy', -1, 1, 2, unit='m'))
    detector.attrs['axes'] = ['xx', 'yy']
    detector.attrs['x_pixel_offset_indices'] = [0]
    detector.attrs['y_pixel_offset_indices'] = [1]
    snx.create_field(
        detector, 'depends_on', sc.scalar('/detector_0/transformations/t1')
    )
    transformations = snx.create_class(detector, 'transformations', NXtransformations)
    log = sc.DataArray(
        sc.array(dims=['time'], values=[1.1, 2.2], unit='m'),
        coords={'time': sc.array(dims=['time'], values=[11, 22], unit='s')},
    )
    log.coords['time'] = sc.epoch(unit='ns') + log.coords['time'].to(unit='ns')
    offset = sc.spatial.translation(value=[1, 2, 3], unit='m')
    vector = sc.vector(value=[0, 0, 1])
    t = log * vector
    t.data = sc.spatial.translations(dims=t.dims, values=t.values, unit=t.unit)
    value = snx.create_class(transformations, 't1', snx.NXlog)
    snx.create_field(value, 'time', log.coords['time'] - sc.epoch(unit='ns'))
    snx.create_field(value, 'value', log.data)
    value.attrs['depends_on'] = '.'
    value.attrs['transformation_type'] = 'translation'
    value.attrs['offset'] = offset.values
    value.attrs['offset_units'] = str(offset.unit)
    value.attrs['vector'] = vector.value

    root = make_group(h5root)
    loaded = root[()]
    result = snx.compute_positions(loaded)
    assert 'position' in result['detector_0']
    assert 'position' not in result['detector_0']['data'].coords
    result = snx.compute_positions(loaded, store_transform='transform')
    assert_identical(result['detector_0']['transform'], t * offset)


def test_compute_positions_warns_if_depends_on_is_dead_link(h5root):
    instrument = snx.create_class(h5root, 'instrument', snx.NXinstrument)
    detector = create_detector(instrument)
    snx.create_field(detector, 'depends_on', sc.scalar('transform'))
    root = make_group(h5root)
    loaded = root[()]
    with pytest.warns(UserWarning, match='depends_on chain references missing node'):
        snx.compute_positions(loaded)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Simon Heybrock
from typing import Dict, Union

import scipp as sc

from .._common import to_plain_index
from .nxobject import NXobject, ScippIndex


class Leaf(NXobject):
    """Base class for "leaf" groups than can be loaded as a dict."""

    def _getitem(
        self, select: ScippIndex
    ) -> Dict[str, Union[sc.Variable, sc.DataArray]]:
        from .nexus_classes import NXtransformations

        index = to_plain_index([], select)
        if index != tuple():
            raise ValueError(f"Cannot select slice when loading {type(self).__name__}")
        content = sc.DataGroup()
        for key, obj in self.items():
            if key == 'depends_on' or isinstance(obj, NXtransformations):
                continue
            content[key] = obj[()]
        return content



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Simon Heybrock
from abc import ABC, abstractmethod
from typing import Dict

from .nxobject import NXobject


class ApplicationDefinition(ABC):
    """Abstract base class that can be subclassed for custom application definitions."""

    @abstractmethod
    def make_strategy(self, group: NXobject) -> type:
        """Return a strategy that ScippNexus should use to load the given group."""
        ...


def make_definition(mapping: Dict[NXobject, type]) -> ApplicationDefinition:
    """
    Create an application definition from a mapping of NeXus classes to strategies.
    """

    class Definition(ApplicationDefinition):
        def make_strategy(self, group: NXobject) -> type:
            return mapping.get(group.nx_class)

    return Definition()
